################################################################################
# UNIT TESTS
################################################################################

import unittest

import numpy as np

from pdstable import PdsTable


class Test_Pds4Table(unittest.TestCase):

    def runTest(self):

        INDEX_PATH = 'test_files/uranus_occultations_index.xml'

        test_table_basic = PdsTable(INDEX_PATH)

        # Test strings
        test_file_names = test_table_basic.column_values['File Name']
        file_name_test_set = np.array([
                'u0201_palomar_508cm_2200nm_radius_equator_egress_1000m.xml',
                'u0201_palomar_508cm_2200nm_radius_equator_egress_100m.xml',
                'u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml',
                'u0201_palomar_508cm_2200nm_radius_equator_ingress_1000m.xml'
                ])
        self.assertTrue(np.all(file_name_test_set == test_file_names[0:4]))

        # Test ints
        test_data_quality_idx = test_table_basic.column_values['Data Quality Index']
        data_quality_idx_test_set = np.array([6, 6, 6, 6])
        self.assertTrue(np.all(data_quality_idx_test_set == test_data_quality_idx[:4]))

        # Test floats
        test_proj_star_diameter = \
                test_table_basic.column_values['Projected Star Diameter']
        proj_star_diameter_test_set = np.array([0.3, 0.3, 0.3, 0.3])
        self.assertTrue(np.all(proj_star_diameter_test_set ==
                               test_proj_star_diameter[:4]))

        # Test times as strings
        test_start_time_strs = test_table_basic.column_values['Start Time UTC']
        start_time_str_test_set = ['2002-07-29T10:03:17.0654Z',
                                   '2002-07-29T10:03:13.4075Z',
                                   '2002-07-29T10:03:15.4524Z',
                                   '2002-07-29T09:48:38.6088Z']
        self.assertEqual(start_time_str_test_set[0], test_start_time_strs[0])
        self.assertEqual(start_time_str_test_set[1], test_start_time_strs[1])
        self.assertEqual(start_time_str_test_set[2], test_start_time_strs[2])
        self.assertEqual(start_time_str_test_set[3], test_start_time_strs[3])

        self.assertTrue(isinstance(test_start_time_strs, np.ndarray))
        self.assertTrue(isinstance(test_start_time_strs[0], np.str_))

        # Test dicts_by_row()
        rowdict = test_table_basic.dicts_by_row()
        for i in range(4):
            self.assertEqual(rowdict[i]['Start Time UTC'], test_start_time_strs[i])

        rowvals = test_table_basic.get_column('Start Time UTC')
        rowmasks = test_table_basic.get_column_mask('Start Time UTC')
        for i in range(10):
            self.assertEqual(rowdict[i]['Start Time UTC'], rowvals[i])
            self.assertFalse(rowmasks[i])

        # Test strings with whitespace preserved
        self.assertEqual(rowdict[22]['Star ID'], 'u0   ')

        ####################################
        # Test times as seconds (floats)
        ####################################

        test_table_secs = PdsTable(INDEX_PATH, times=['Start Time UTC'])

        test_start_times = test_table_secs.column_values['Start Time UTC']
        start_time_test_set = np.array([81209029.0654, 81209025.4075,
                                        81209027.4524, 81208150.6088])
        self.assertTrue(np.all(start_time_test_set == test_start_times[:4]))
        self.assertTrue(isinstance(start_time_test_set, np.ndarray))

        # Test dicts_by_row()
        rowdict = test_table_secs.dicts_by_row()
        for i in range(4):
            self.assertEqual(rowdict[i]['Start Time UTC'], start_time_test_set[i])

        rowvals = test_table_secs.get_column('Start Time UTC')
        rowmask = test_table_secs.get_column_mask('Start Time UTC')
        for i in range(10):
            self.assertEqual(rowdict[i]['Start Time UTC'], rowvals[i])
            self.assertFalse(rowmask[i])

        ####################################
        # Row lookups
        ####################################
        # File Specification
        self.assertEqual(test_table_basic.filespec_column_index(), 2)
        # Bundle Name
        self.assertEqual(test_table_basic.volume_column_index(), 4)
        # File Specification
        self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
                '',
                r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
                r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), 2)
        self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
                '',
                r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
                r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), [2])
        self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
                '',
                r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
                r'\u0201_palomar_508cm_2200nm_radius_equator_ingress_1000m.xml'), 3)
        self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
                '',
                r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
                r'\u0201_palomar_508cm_2200nm_radius_equator_ingress_1000m.xml'), [3])
        # Bundle Name & File Specification
        self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
                'uranus_occ_u0_kao_91cm',
                r'2021-04-uranus-redelivery\uranus_occ_u0_kao_91cm\data\global' +
                r'\u0_kao_91cm_734nm_radius_equator_egress_1000m.xml'), 21)
        self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
                'uranus_occ_u0_kao_91cm',
                r'2021-04-uranus-redelivery\uranus_occ_u0_kao_91cm\data\global' +
                r'\u0_kao_91cm_734nm_radius_equator_egress_1000m.xml'), [21])

        ####################################
        # Row ranges
        ####################################

        partial_table = PdsTable(INDEX_PATH, row_range=(2, 4))
        self.assertEqual(partial_table.rows, 2)

        self.assertEqual(partial_table.filespec_column_index(), 2)
        self.assertEqual(partial_table.volume_column_index(), 4)

        self.assertEqual(partial_table.find_row_index_by_volume_filespec(
                '',
                r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
                r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), 0)
        self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
                '',
                r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
                r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), [0])

        self.assertEqual(partial_table.find_row_index_by_volume_filespec(
                '',
                r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
                r'\u0201_palomar_508cm_2200nm_radius_equator_ingress_1000m.xml'), 1)
        self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
                '',
                r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
                r'\u0201_palomar_508cm_2200nm_radius_equator_ingress_1000m.xml'), [1])

        self.assertEqual(partial_table.find_row_index_by_volume_filespec(
                'uranus_occ_u0201_palomar_508cm',
                r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
                r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), 0)
        self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
                'uranus_occ_u0201_palomar_508cm',
                r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
                r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), [0])

        ####################################
        # PdsLabel input option
        ####################################
        # For PDS4, we store the label dictionary in .lable instead of pdsparser.PdsLabel
        # instance, therefore we use "==" here instead of "is"
        test = PdsTable(INDEX_PATH, label_contents=partial_table.pdslabel)
        self.assertTrue(test.pdslabel == partial_table.pdslabel)

        # PDS4 TODO: Add tests for invalids & replacements



################################################################################
# UNIT TESTS
################################################################################

from pathlib import Path
import unittest
import warnings

import numpy as np

from pdstable import PdsTable


class Test_Pds3Table(unittest.TestCase):

    def runTest(self):

        # Testing different values parsed correctly...
        INDEX_PATH = "test_files/cassini_iss_index.lbl"
        EDITED_INDEX_PATH = "test_files/cassini_iss_index_edited.lbl"

        test_table_some_cols = PdsTable(INDEX_PATH, columns=['FILE_NAME', 'START_TIME'])

        self.assertEqual(test_table_some_cols.columns, 2)
        self.assertEqual(test_table_some_cols.all_columns, 118)
        self.assertEqual(len(test_table_some_cols.column_values), 2)
        self.assertEqual(len(test_table_some_cols.column_masks), 2)
        self.assertEqual(len(test_table_some_cols.column_info_list), 118)
        self.assertEqual(len(test_table_some_cols.column_info_dict), 118)

        test_table_basic = PdsTable(INDEX_PATH)

        cwd = Path.cwd()

        self.assertEqual(test_table_basic.label_file_name, 'cassini_iss_index.lbl')
        self.assertEqual(test_table_basic.label_file_path,
                         cwd / 'test_files/cassini_iss_index.lbl')
        self.assertEqual(test_table_basic.table_file_name, 'cassini_iss_index.tab')
        self.assertEqual(test_table_basic.table_file_path,
                         cwd / 'test_files/cassini_iss_index.tab')

        self.assertEqual(test_table_basic.is_pds4, False)
        self.assertEqual(test_table_basic.encoding, 'latin-1')
        self.assertEqual(test_table_basic.first, 0)
        self.assertEqual(test_table_basic.rows, 4575)
        self.assertEqual(test_table_basic.columns, 118)
        self.assertEqual(test_table_basic.header_bytes, 0)
        self.assertEqual(test_table_basic.row_bytes, 3057)
        self.assertEqual(test_table_basic.info.row_bytes, 3057)  # Deprecated
        self.assertEqual(test_table_basic.fixed_length_row, True)
        self.assertEqual(test_table_basic.field_delimiter, None)
        self.assertEqual(len(test_table_basic.column_info_list), 118)
        self.assertEqual(len(test_table_basic.column_info_dict), 118)
        self.assertEqual(test_table_basic.dtype0['crlf'], ('|S2', 3055))

        # Test strings
        test_file_names = test_table_basic.column_values['FILE_NAME']
        file_name_test_set = np.array(['N1573186009_1.IMG',
                                       'W1573186009_1.IMG',
                                       'N1573186041_1.IMG',
                                       'W1573186041_1.IMG'])
        self.assertTrue(np.all(file_name_test_set == test_file_names[0:4]))

        # Test floats
        test_cbody_dists = test_table_basic.column_values['CENTRAL_BODY_DISTANCE']
        cent_body_dist_test_set = np.array([2869736.9, 2869736, 2869707,
                                            2869706.9])
        self.assertTrue(np.all(cent_body_dist_test_set == test_cbody_dists[0:4]))

        # test vectors
        test_sc_vels = test_table_basic.column_values['SC_TARGET_VELOCITY_VECTOR']
        sc_vels_test_set = np.array([[1.2223705, -1.1418157, -0.055303727],
                                     [1.2223749, -1.1418146, -0.055303917],
                                     [1.2225166, -1.1417793, -0.055309978],
                                     [1.2225173, -1.1417791, -0.055310007]])
        self.assertTrue(np.all(sc_vels_test_set == test_sc_vels[0:4]))

        # Test times as strings
        test_start_time_strs = test_table_basic.column_values['START_TIME']
        start_time_str_test_set = ['2007-312T03:31:12.392',
                                   '2007-312T03:31:14.372',
                                   '2007-312T03:31:45.832',
                                   '2007-312T03:31:46.132']
        self.assertEqual(start_time_str_test_set[0], test_start_time_strs[0])
        self.assertEqual(start_time_str_test_set[1], test_start_time_strs[1])
        self.assertEqual(start_time_str_test_set[2], test_start_time_strs[2])
        self.assertEqual(start_time_str_test_set[3], test_start_time_strs[3])

        self.assertTrue(isinstance(test_start_time_strs, np.ndarray))
        self.assertTrue(isinstance(test_start_time_strs[0], np.str_))

        # Test dicts_by_row()
        rowdict = test_table_basic.dicts_by_row()
        for i in range(4):
            self.assertEqual(rowdict[i]["START_TIME"], test_start_time_strs[i])

        rowvals = test_table_basic.get_column("START_TIME")
        self.assertTrue(rowvals is test_table_basic.column_values["START_TIME"])
        rowmasks = test_table_basic.get_column_mask("START_TIME")
        self.assertTrue(rowmasks is test_table_basic.column_masks["START_TIME"])
        for i in range(10):
            self.assertEqual(rowdict[i]["START_TIME"], rowvals[i])
            self.assertFalse(rowmasks[i])

        ####################################
        # Test string stripping
        ####################################

        self.assertEqual(test_table_basic.column_values['FILE_NAME'][0],
                         'N1573186009_1.IMG')

        test_table_no_strip = PdsTable(INDEX_PATH, columns=['FILE_NAME'],
                                       nostrip=['FILE_NAME'])
        self.assertEqual(test_table_no_strip.column_values['FILE_NAME'][0],
                         'N1573186009_1.IMG     ')

        ####################################
        # Test times as seconds (floats)
        ####################################

        test_table_secs = PdsTable(INDEX_PATH, times=['START_TIME'])

        test_start_times = test_table_secs.column_values['START_TIME']
        start_time_test_set = np.array([247764705.392, 247764707.372,
                                        247764738.832, 247764739.132])
        self.assertTrue(np.all(start_time_test_set == test_start_times[0:4]))
        self.assertTrue(isinstance(start_time_test_set, np.ndarray))

        # Test dicts_by_row()
        rowdict = test_table_secs.dicts_by_row()
        for i in range(4):
            self.assertEqual(rowdict[i]["START_TIME"], start_time_test_set[i])

        rowvals = test_table_secs.get_column("START_TIME")
        rowmask = test_table_secs.get_column_mask("START_TIME")
        for i in range(10):
            self.assertEqual(rowdict[i]["START_TIME"], rowvals[i])
            self.assertFalse(rowmask[i])

        ####################################
        # Invalids
        ####################################

        test_table = PdsTable(INDEX_PATH, times=['START_TIME'],
                              invalid={'default': [-1.e32, -2147483648]})

        rowdict = test_table_secs.dicts_by_row()
        for key in test_table.get_keys():
            if key.endswith('_mask'):
                continue

            rowmasks = test_table_secs.get_column_mask(key)
            self.assertFalse(np.any(rowmasks))
            self.assertTrue(isinstance(rowmasks, np.ndarray))

        results = {
            'BIAS_STRIP_MEAN': 0,
            'COMMAND_SEQUENCE_NUMBER': 0,
            'DARK_STRIP_MEAN': 0,
            'DETECTOR_TEMPERATURE': 0,
            'ELECTRONICS_BIAS': 0,
            'EXPECTED_PACKETS': 0,
            'EXPOSURE_DURATION': 0,
            'FILTER_TEMPERATURE': 0,
            'INSTRUMENT_DATA_RATE': 0,
            'INST_CMPRS_RATIO': 0,
            'MISSING_LINES': 1102,
            'ORDER_NUMBER': 0,
            'PARALLEL_CLOCK_VOLTAGE_INDEX': 0,
            'PREPARE_CYCLE_INDEX': 0,
            'READOUT_CYCLE_INDEX': 0,
            'RECEIVED_PACKETS': 0,
            'SENSOR_HEAD_ELEC_TEMPERATURE': 0,
            'SEQUENCE_NUMBER': 0,
            'SPACECRAFT_CLOCK_CNT_PARTITION': 0,
            'START_TIME': 0,
            'CENTRAL_BODY_DISTANCE': 0,
            'DECLINATION': 0,
            'EMISSION_ANGLE': 1563,
            'INCIDENCE_ANGLE': 1563,
            'LOWER_LEFT_LATITUDE': 3426,
            'LOWER_LEFT_LONGITUDE': 3426,
            'LOWER_RIGHT_LATITUDE': 3279,
            'LOWER_RIGHT_LONGITUDE': 3279,
            'MAXIMUM_RING_RADIUS': 110,
            'MINIMUM_RING_RADIUS': 110,
            'NORTH_AZIMUTH_CLOCK_ANGLE': 1563,
            'PHASE_ANGLE': 236,
            'PIXEL_SCALE': 236,
            'RIGHT_ASCENSION': 0,
            'RING_CENTER_LATITUDE': 3612,
            'RING_CENTER_LONGITUDE': 3612,
            'RING_EMISSION_ANGLE': 3612,
            'RING_INCIDENCE_ANGLE': 3612,
            'SUB_SOLAR_LATITUDE': 236,
            'SUB_SOLAR_LONGITUDE': 236,
            'SUB_SPACECRAFT_LATITUDE': 236,
            'SUB_SPACECRAFT_LONGITUDE': 236,
            'CENTER_LATITUDE': 1563,
            'CENTER_LONGITUDE': 1563,
            'TARGET_DISTANCE': 236,
            'TARGET_EASTERNMOST_LONGITUDE': 1006,
            'TARGET_NORTHERNMOST_LATITUDE': 1006,
            'TARGET_SOUTHERNMOST_LATITUDE': 1006,
            'TARGET_WESTERNMOST_LONGITUDE': 1006,
            'TWIST_ANGLE': 0,
            'UPPER_LEFT_LATITUDE': 3144,
            'UPPER_LEFT_LONGITUDE': 3144,
            'UPPER_RIGHT_LATITUDE': 3102,
            'UPPER_RIGHT_LONGITUDE': 3102,
        }

        rowdict = test_table_secs.dicts_by_row()
        for key in test_table_secs.get_keys():
            if key.endswith('_mask'):
                continue

            rowvals = test_table.get_column(key)
            if np.shape(rowvals[0]) != ():
                continue

            rowmask = test_table.get_column_mask(key)
            if rowvals.dtype.kind == 'f':
                countv = np.sum(rowvals == -1.e32)
                countm = np.sum(rowmask)
                self.assertEqual(countv, countm)
                self.assertEqual(countv, results[key])

            elif rowvals.dtype.kind == 'i':
                countv = np.sum(rowvals == -2147483648)
                countm = np.sum(rowmask)
                self.assertEqual(countv, countm)
                self.assertEqual(countv, results[key])

            else:
                self.assertEqual(np.sum(rowmask), 0)

        # 22.5 is a common value in column BIAS_STRIP_MEAN
        test_table = PdsTable(INDEX_PATH, times=['START_TIME'],
                              invalid={'default': [-1.e32, -2147483648, 22.5]})

        key = 'BIAS_STRIP_MEAN'
        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 511)

        test_table = PdsTable(INDEX_PATH, times=['START_TIME'],
                              invalid={'default': [-1.e32, -2147483648], key: 22.5})

        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 511)

        ####################################
        # Replacements
        ####################################

        # Replacement as a number
        key = 'BIAS_STRIP_MEAN'
        test_table = PdsTable(INDEX_PATH, times=['START_TIME'],
                              invalid={'default': [-1.e32, -2147483648]},
                              replacements={key: {22.5: -1.e32}})
        rowvals = test_table.get_column(key)
        self.assertEqual(np.sum(rowvals == 22.5), 0)
        self.assertEqual(np.sum(rowvals == -1.e32), 511)

        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 511)

        # Replacement as a string
        test_table = PdsTable(INDEX_PATH, times=['START_TIME'],
                              invalid={'default': [-1.e32, -2147483648]},
                              replacements={key: {'       22.5': '     -1.e32'}})
        rowvals = test_table.get_column(key)
        self.assertEqual(np.sum(rowvals == 22.5), 0)
        self.assertEqual(np.sum(rowvals == -1.e32), 511)

        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 511)

        # Replacement via a callback
        def test_callback_as_str(arg):
            if arg.strip() == '22.5':
                return '-1e32'
            return arg

        test_table = PdsTable(INDEX_PATH, times=['START_TIME'],
                              invalid={'default': [-1.e32, -2147483648]},
                              callbacks={key: test_callback_as_str})
        rowvals = test_table.get_column(key)
        self.assertEqual(np.sum(rowvals == 22.5), 0)
        self.assertEqual(np.sum(rowvals == -1.e32), 511)

        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 511)

        # Replacement via an ASCII byte string callback
        def test_callback_as_bytes(arg):
            if arg.strip() == b'22.5':
                return b'-1e32'
            return arg

        test_table = PdsTable(INDEX_PATH, times=['START_TIME'],
                              invalid={'default': [-1.e32, -2147483648]},
                              callbacks={key: test_callback_as_bytes}, ascii=True)
        rowvals = test_table.get_column(key)
        self.assertEqual(np.sum(rowvals == 22.5), 0)
        self.assertEqual(np.sum(rowvals == -1.e32), 511)

        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 511)

        ####################################
        # "UNK" values replace 22.5 in BIAS_STRIP_MEAN
        # "NULL" in second row of table for CALIBRATION_LAMP_STATE_FLAG
        # "UNK" in the first row for IMAGE_MID_TIME
        # Label says INVALID_CONSTANT = 19.5 for DARK_STRIP_MEAN
        # Label says VALID_RANGE = (2,3) for INST_CMPRS_RATE
        # Manually disallow negative values for FILTER_TEMPERATURE
        # Every value of INSTRUMENT_DATA_RATE is exactly 182.783997 except one.
        ####################################

    #     print('')
    #     print('Two UserWarnings should follow...')
    #     print('')
    #     print('25 illegally formatted float values in column BIAS_STRIP_MEAN; ' +
    #           'first example is "UNK"')
    #     print('Illegally formatted time value in column IMAGE_MID_TIME: UNK')
    #     print('')

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            test_table = PdsTable(EDITED_INDEX_PATH, times=['IMAGE_MID_TIME'],
                                  invalid={'default': [-1.e32, -2147483648]},
                                  replacements={'INSTRUMENT_DATA_RATE':
                                                {182.783997: 1.}},
                                  valid_ranges={'FILTER_TEMPERATURE': [0., 1.e99]})

    #     print('')
    #     print('')

        image_mid_time = test_table.get_column('IMAGE_MID_TIME')
        bias_strip_mean = test_table.get_column('BIAS_STRIP_MEAN')
        self.assertEqual(type(image_mid_time), list)
        self.assertEqual(type(image_mid_time), list)

        self.assertEqual(image_mid_time[0], 'UNK')
        self.assertEqual(type(image_mid_time[0]), str)
        for value in image_mid_time[1:]:
            self.assertTrue(isinstance(value, float))

        for value in bias_strip_mean:
            self.assertTrue((value == 'UNK') or isinstance(value, float))

        dark_strip_mean = test_table.get_column('DARK_STRIP_MEAN')
        dsm_mask = test_table.get_column_mask('DARK_STRIP_MEAN')
        for (value, flag) in zip(dark_strip_mean, dsm_mask):
            self.assertTrue(flag == (value == 19.5))

        inst_cmprs_rate = test_table.get_column('INST_CMPRS_RATE')
        icr_mask = test_table.get_column_mask('INST_CMPRS_RATE')
        for (value, flag) in zip(inst_cmprs_rate, icr_mask):
            self.assertTrue(flag[0] == (value[0] < 2 or value[0] > 3))
            self.assertTrue(flag[1] == (value[1] < 2 or value[1] > 3))

        filter_temperature = test_table.get_column('FILTER_TEMPERATURE')
        ft_mask = test_table.get_column_mask('FILTER_TEMPERATURE')
        for (value, flag) in zip(filter_temperature, ft_mask):
            self.assertTrue(flag == (value < 0.))

        instrument_data_rate = test_table.get_column('INSTRUMENT_DATA_RATE')
        idr_mask = test_table.get_column_mask('INSTRUMENT_DATA_RATE')
        self.assertTrue(np.sum(instrument_data_rate == 1.) == 99)
        self.assertTrue(np.all(instrument_data_rate != 182.783997))
        self.assertTrue(not np.any(idr_mask))

        ####################################
        # Row lookups
        ####################################

        self.assertEqual(test_table_basic.filespec_column_index(), 1)
        self.assertEqual(test_table_basic.volume_column_index(), 2)

        self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
                '', 'data/1573186009_1573197826/N1573186041_1.IMG'), 2)
        self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
                '', 'data/1573186009_1573197826/N1573186041_1.IMG'), [2])

        self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
                'COISS_2039', 'data/1573186009_1573197826/N1573186041_1.IMG'), 2)
        self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
                'COISS_2039', 'data/1573186009_1573197826/N1573186041_1.IMG'), [2])

        self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
                'coiss_2039', 'data/1573186009_1573197826/N1573186041_1.IMG'), 2)
        self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
                'coiss_2039', 'data/1573186009_1573197826/N1573186041_1.IMG'), [2])

        ####################################
        # Row ranges
        ####################################

        partial_table = PdsTable(INDEX_PATH, row_range=(2, 4))
        self.assertEqual(partial_table.rows, 2)

        self.assertEqual(partial_table.filespec_column_index(), 1)
        self.assertEqual(partial_table.volume_column_index(), 2)

        self.assertEqual(partial_table.find_row_index_by_volume_filespec(
                '', 'data/1573186009_1573197826/N1573186041_1.IMG'), 0)
        self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
                '', 'data/1573186009_1573197826/N1573186041_1.IMG'), [0])

        self.assertEqual(partial_table.find_row_index_by_volume_filespec(
                'COISS_2039', 'data/1573186009_1573197826/N1573186041_1.IMG'), 0)
        self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
                'COISS_2039', 'data/1573186009_1573197826/N1573186041_1.IMG'), [0])

        self.assertEqual(partial_table.find_row_index_by_volume_filespec(
                'coiss_2039', 'data/1573186009_1573197826/N1573186041_1.IMG'), 0)
        self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
                'coiss_2039', 'data/1573186009_1573197826/N1573186041_1.IMG'), [0])

        ####################################
        # PdsLabel input option
        ####################################

        test = PdsTable(INDEX_PATH, label_contents=partial_table.pdslabel)
        self.assertIs(test.pdslabel, partial_table.pdslabel)

        test = PdsTable(INDEX_PATH, label_contents=partial_table.pdslabel.content)
        self.assertIsNot(test.pdslabel, partial_table.pdslabel)
        self.assertEqual(test.pdslabel.content, partial_table.pdslabel.content)

        ####################################
        # Other PdsTable options
        ####################################

        # table_file
        self.assertRaises(ValueError, PdsTable, INDEX_PATH, table_file=1)

        # row_range
        self.assertRaises(ValueError, PdsTable, EDITED_INDEX_PATH, row_range=(99, 98))
        self.assertRaises(ValueError, PdsTable, EDITED_INDEX_PATH, row_range=(99, 101))
        PdsTable(EDITED_INDEX_PATH, row_range=(99, 100))

        # table_callback
        def test_callback(b):
            return [s.replace(b'2007', b'2009') for s in b]

        test_table_callback = PdsTable(INDEX_PATH, table_callback=test_callback)

        test_start_time_strs = test_table_callback.column_values['START_TIME']
        start_time_str_test_set = ['2009-312T03:31:12.392',
                                   '2009-312T03:31:14.372',
                                   '2009-312T03:31:45.832',
                                   '2009-312T03:31:46.132']
        self.assertEqual(start_time_str_test_set[0], test_start_time_strs[0])
        self.assertEqual(start_time_str_test_set[1], test_start_time_strs[1])
        self.assertEqual(start_time_str_test_set[2], test_start_time_strs[2])
        self.assertEqual(start_time_str_test_set[3], test_start_time_strs[3])

        self.assertEqual(test_table_basic.column_info_dict['START_TIME'].colno, 63)
        self.assertEqual(test_table_basic.column_info_dict['START_TIME'].start_byte, 1504)
        self.assertEqual(test_table_basic.column_info_dict['START_TIME'].bytes, 22)
        self.assertEqual(test_table_basic.column_info_dict['START_TIME'].items, 1)
        self.assertEqual(test_table_basic.column_info_dict['START_TIME'].item_bytes, 22)
        self.assertEqual(test_table_basic.column_info_dict['START_TIME'].item_offset, 22)
        self.assertEqual(test_table_basic.column_info_dict['START_TIME'].dtype0,
                         ('S22', 1503))
        self.assertEqual(test_table_basic.column_info_dict['START_TIME'].dtype1, None)

        ####################################
        # Bad label file
        ####################################

        self.assertRaises(ValueError, PdsTable,
                          EDITED_INDEX_PATH.replace('.lbl', '_bad_rows.lbl'))
        self.assertRaises(IOError, PdsTable,
                          EDITED_INDEX_PATH.replace('.lbl', '_not_fixed.lbl'))
        # This next call needs to test that a warning is raised
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            PdsTable(EDITED_INDEX_PATH.replace('.lbl', '_not_simple.lbl'))
            self.assertEqual(w[0].category, UserWarning)
            self.assertIn('Simple Pointer', str(w[0].message))
        warnings.resetwarnings()
        self.assertRaises(IOError, PdsTable,
                          EDITED_INDEX_PATH.replace('.lbl', '_no_ptr.lbl'))
        self.assertRaises(IOError, PdsTable,
                          EDITED_INDEX_PATH.replace('.lbl', '_not_ascii.lbl'))
        self.assertRaises(ValueError, PdsTable,
                          EDITED_INDEX_PATH.replace('.lbl', '_dup_col.lbl'))
        self.assertRaises(IOError, PdsTable,
                          EDITED_INDEX_PATH.replace('.lbl', '_bad_data_type.lbl'))



################################################################################
# UNIT TESTS FOR UTILS.PY
################################################################################

import unittest

import numpy as np

from pdstable.utils import (is_pds4_label, tai_from_iso, int_from_base2,
                            int_from_base8, int_from_base16, lowercase_value)


class TestUtils(unittest.TestCase):
    """Test cases for utility functions in utils.py."""

    def test_is_pds4_label(self):
        """Test is_pds4_label function with various file extensions."""

        # Test PDS4 label extensions
        self.assertTrue(is_pds4_label('data.xml'))
        self.assertTrue(is_pds4_label('data.lblx'))
        self.assertTrue(is_pds4_label('/path/to/file.xml'))
        self.assertTrue(is_pds4_label('C:\\path\\to\\file.lblx'))

        # Test non-PDS4 label extensions
        self.assertFalse(is_pds4_label('data.lbl'))
        self.assertFalse(is_pds4_label('data.tab'))
        self.assertFalse(is_pds4_label('data.txt'))
        self.assertFalse(is_pds4_label('data'))
        self.assertFalse(is_pds4_label(''))

        # Test edge cases - the function only checks if the string ends with the
        # extensions
        self.assertTrue(is_pds4_label('.xml'))  # Just extension
        self.assertFalse(is_pds4_label('xml'))   # No dot
        self.assertFalse(is_pds4_label('data.xml.bak'))  # Multiple extensions doesn't
        # end with .xml

        # Test case sensitivity - the function is case-sensitive
        self.assertFalse(is_pds4_label('data.XML'))  # Uppercase extension
        self.assertFalse(is_pds4_label('data.LBLX'))  # Uppercase extension

    def test_tai_from_iso(self):
        """Test tai_from_iso function with various ISO time strings."""

        # Test basic ISO time strings
        # Note: These are example values - actual TAI values may differ
        # We're testing that the function doesn't crash and returns a number

        # Test various ISO formats
        result1 = tai_from_iso('2007-312T03:31:12.392')
        self.assertIsInstance(result1, (int, float))  # Can be either int or float
        self.assertGreater(result1, 0)

        result2 = tai_from_iso('2007-01-01T00:00:00')
        self.assertIsInstance(result2, (int, float))  # Can be either int or float
        self.assertGreater(result2, 0)

        result3 = tai_from_iso('2007-001T00:00:00.000')
        self.assertIsInstance(result3, (int, float))  # Can be either int or float
        self.assertGreater(result3, 0)

        # Test that strip=True works (removes whitespace)
        result4 = tai_from_iso('  2007-312T03:31:12.392  ')
        self.assertIsInstance(result4, (int, float))  # Can be either int or float
        self.assertEqual(result4, result1)

        # Test edge cases
        with self.assertRaises(Exception):  # Should raise some kind of error
            tai_from_iso('invalid_time')

        with self.assertRaises(Exception):
            tai_from_iso('')

        with self.assertRaises(Exception):
            tai_from_iso('2007-13-45T25:70:99')  # Invalid date/time

    def test_int_from_base2(self):
        """Test int_from_base2 function with various binary strings."""

        # Test valid binary strings
        self.assertEqual(int_from_base2('0'), 0)
        self.assertEqual(int_from_base2('1'), 1)
        self.assertEqual(int_from_base2('10'), 2)
        self.assertEqual(int_from_base2('11'), 3)
        self.assertEqual(int_from_base2('100'), 4)
        self.assertEqual(int_from_base2('101'), 5)
        self.assertEqual(int_from_base2('110'), 6)
        self.assertEqual(int_from_base2('111'), 7)
        self.assertEqual(int_from_base2('1000'), 8)
        self.assertEqual(int_from_base2('1001'), 9)
        self.assertEqual(int_from_base2('1010'), 10)

        # Test longer binary strings
        self.assertEqual(int_from_base2('10101010'), 170)
        self.assertEqual(int_from_base2('11111111'), 255)
        self.assertEqual(int_from_base2('100000000'), 256)

        # Test edge cases
        self.assertEqual(int_from_base2('0000'), 0)
        self.assertEqual(int_from_base2('0001'), 1)

        # Test error cases
        with self.assertRaises(ValueError):
            int_from_base2('2')  # Invalid binary digit

        with self.assertRaises(ValueError):
            int_from_base2('10a')  # Invalid binary digit

        with self.assertRaises(ValueError):
            int_from_base2('')  # Empty string

        with self.assertRaises(ValueError):
            int_from_base2('abc')  # No valid digits

    def test_int_from_base8(self):
        """Test int_from_base8 function with various octal strings."""

        # Test valid octal strings
        self.assertEqual(int_from_base8('0'), 0)
        self.assertEqual(int_from_base8('1'), 1)
        self.assertEqual(int_from_base8('7'), 7)
        self.assertEqual(int_from_base8('10'), 8)
        self.assertEqual(int_from_base8('11'), 9)
        self.assertEqual(int_from_base8('17'), 15)
        self.assertEqual(int_from_base8('20'), 16)
        self.assertEqual(int_from_base8('77'), 63)
        self.assertEqual(int_from_base8('100'), 64)
        self.assertEqual(int_from_base8('777'), 511)

        # Test edge cases
        self.assertEqual(int_from_base8('0000'), 0)
        self.assertEqual(int_from_base8('0001'), 1)

        # Test error cases
        with self.assertRaises(ValueError):
            int_from_base8('8')  # Invalid octal digit

        with self.assertRaises(ValueError):
            int_from_base8('18')  # Invalid octal digit

        with self.assertRaises(ValueError):
            int_from_base8('')  # Empty string

        with self.assertRaises(ValueError):
            int_from_base8('abc')  # No valid digits

    def test_int_from_base16(self):
        """Test int_from_base16 function with various hexadecimal strings."""

        # Test valid hex strings
        self.assertEqual(int_from_base16('0'), 0)
        self.assertEqual(int_from_base16('1'), 1)
        self.assertEqual(int_from_base16('9'), 9)
        self.assertEqual(int_from_base16('a'), 10)
        self.assertEqual(int_from_base16('A'), 10)
        self.assertEqual(int_from_base16('f'), 15)
        self.assertEqual(int_from_base16('F'), 15)
        self.assertEqual(int_from_base16('10'), 16)
        self.assertEqual(int_from_base16('1f'), 31)
        self.assertEqual(int_from_base16('20'), 32)
        self.assertEqual(int_from_base16('ff'), 255)
        self.assertEqual(int_from_base16('FF'), 255)
        self.assertEqual(int_from_base16('100'), 256)

        # Test longer hex strings
        self.assertEqual(int_from_base16('ffff'), 65535)
        self.assertEqual(int_from_base16('10000'), 65536)

        # Test edge cases
        self.assertEqual(int_from_base16('0000'), 0)
        self.assertEqual(int_from_base16('0001'), 1)

        # Test error cases
        with self.assertRaises(ValueError):
            int_from_base16('g')  # Invalid hex digit

        with self.assertRaises(ValueError):
            int_from_base16('1g')  # Invalid hex digit

        with self.assertRaises(ValueError):
            int_from_base16('')  # Empty string

        with self.assertRaises(ValueError):
            int_from_base16('xyz')  # No valid digits

    def test_lowercase_value_string(self):
        """Test lowercase_value function with string inputs."""

        # Test basic string conversion
        self.assertEqual(lowercase_value('Hello'), 'hello')
        self.assertEqual(lowercase_value('WORLD'), 'world')
        self.assertEqual(lowercase_value('MiXeD'), 'mixed')
        self.assertEqual(lowercase_value(''), '')
        self.assertEqual(lowercase_value('123'), '123')
        self.assertEqual(lowercase_value('!@#'), '!@#')

        # Test unicode strings
        self.assertEqual(lowercase_value('HÉLLÖ'), 'héllö')
        self.assertEqual(lowercase_value('ÑOÑO'), 'ñoño')

    def test_lowercase_value_tuple(self):
        """Test lowercase_value function with tuple inputs."""

        # Test tuple with strings
        input_tuple = ('Hello', 'WORLD', 'MiXeD')
        expected = ('hello', 'world', 'mixed')  # Function returns a list, not tuple
        result = lowercase_value(input_tuple)
        self.assertEqual(result, expected)

        # Test tuple with mixed types
        input_tuple = ('Hello', 123, 'WORLD', 45.67, 'MiXeD')
        expected = ('hello', 123, 'world', 45.67, 'mixed')  # Function returns a list
        result = lowercase_value(input_tuple)
        self.assertEqual(result, expected)

        # Test empty tuple
        self.assertEqual(lowercase_value(()), ())  # Function returns empty list

        # Test single element tuple
        self.assertEqual(lowercase_value(('Hello',)), ('hello',))  # Function returns list

        # Test nested tuples (should not be processed)
        input_tuple = ('Hello', ('WORLD', 'MiXeD'))
        expected = ('hello', ('WORLD', 'MiXeD'))  # Nested tuple unchanged, but result is
        # tuple
        result = lowercase_value(input_tuple)
        self.assertEqual(result, expected)

    def test_lowercase_value_numpy_array(self):
        """Test lowercase_value function with numpy array inputs."""

        # Test numpy array with strings
        input_array = np.array(['Hello', 'WORLD', 'MiXeD'])
        expected = np.array(['hello', 'world', 'mixed'])
        result = lowercase_value(input_array)
        np.testing.assert_array_equal(result, expected)

        # Test numpy array with mixed types
        input_array = np.array(['Hello', 'WORLD', 'MiXeD'], dtype=object)
        expected = np.array(['hello', 'world', 'mixed'], dtype=object)
        result = lowercase_value(input_array)
        np.testing.assert_array_equal(result, expected)

        # Test empty numpy array
        input_array = np.array([])
        result = lowercase_value(input_array)
        np.testing.assert_array_equal(result, input_array)

        # Test single element numpy array
        input_array = np.array(['Hello'])
        expected = np.array(['hello'])
        result = lowercase_value(input_array)
        np.testing.assert_array_equal(result, expected)

        # Test 2D numpy array - the function only processes 1D arrays
        # For 2D arrays, it doesn't process the nested elements
        input_array = np.array([['Hello', 'WORLD'], ['MiXeD', 'Case']])
        result = lowercase_value(input_array)
        # The function doesn't process 2D arrays recursively, so the result should be
        # unchanged
        np.testing.assert_array_equal(result, input_array)

    def test_lowercase_value_other_types(self):
        """Test lowercase_value function with non-string types."""

        # Test numeric types (should be unchanged)
        self.assertEqual(lowercase_value(123), 123)
        self.assertEqual(lowercase_value(45.67), 45.67)
        self.assertEqual(lowercase_value(True), True)
        self.assertEqual(lowercase_value(False), False)
        self.assertEqual(lowercase_value(None), None)

        # Test list (should be unchanged)
        input_list = ['Hello', 'WORLD']
        result = lowercase_value(input_list)
        self.assertEqual(result, input_list)  # Lists are not processed

        # Test dict (should be unchanged)
        input_dict = {'key': 'value'}
        result = lowercase_value(input_dict)
        self.assertEqual(result, input_dict)  # Dicts are not processed

        # Test custom object (should be unchanged)
        class CustomObject:
            pass
        custom_obj = CustomObject()
        result = lowercase_value(custom_obj)
        self.assertEqual(result, custom_obj)

    def test_lowercase_value_edge_cases(self):
        """Test lowercase_value function with edge cases."""

        # Test very long strings
        long_string = 'A' * 1000
        result = lowercase_value(long_string)
        self.assertEqual(result, 'a' * 1000)

        # Test strings with special characters
        special_string = '!@#$%^&*()_+-=[]{}|;:\',./<>?'
        result = lowercase_value(special_string)
        self.assertEqual(result, special_string)  # Special chars unchanged

        # Test strings with numbers
        mixed_string = 'Hello123World456'
        result = lowercase_value(mixed_string)
        self.assertEqual(result, 'hello123world456')

        # Test strings with spaces
        spaced_string = '  Hello  World  '
        result = lowercase_value(spaced_string)
        self.assertEqual(result, '  hello  world  ')

    def test_lowercase_value_performance(self):
        """Test lowercase_value function performance with large inputs."""

        # Test with large numpy array
        large_array = np.array(['Hello'] * 10000)
        result = lowercase_value(large_array)
        self.assertEqual(len(result), 10000)
        self.assertEqual(result[0], 'hello')
        self.assertEqual(result[-1], 'hello')

        # Test with large tuple
        large_tuple = ('Hello',) * 1000
        result = lowercase_value(large_tuple)
        self.assertEqual(len(result), 1000)
        self.assertEqual(result[0], 'hello')
        self.assertEqual(result[-1], 'hello')



################################################################################
# pdstable/pds4table.py
# Pds4TableInfo and Pds4ColumnInfo
################################################################################

from collections import defaultdict
import numbers
import re

from pds4_tools.reader.label_objects import Label

from .pdsxtable import PdsColumnInfo, PdsTableInfo
from .utils import (int_from_base2,
                    int_from_base8,
                    int_from_base16,
                    string_collapsed,
                    tai_from_iso,
                    STRING_TYPES)


# A list of possible table column names that store the bundle name.
PDS4_BUNDLE_COLNAMES_lc = (
    'bundle_name',
    'bundle name',
    'bundle',
)

# A list of possible table column names (unique) that store the path of the file spec or
# the file name. This will be used to get the index of the column containing file
# specification path or name. The order of the values matters in this variable.
# Since there is no standard name in PDS4 for this column (if it's even present),
# we look for these fields that are created by the RMS Node's
# pds4_create_xml_index tool.
PDS4_FILE_SPECIFICATION_COLUMN_NAMES_lc = (
    'file_specification',
    'file specification',
    'file_specification_name',
    'file specification name',
    'filepath',
    'filename',
    'file name',
    'file_name',
)

# The mapping of a product tag to its corresponding file area tags
# The key is a product component tag, and the value is its corresponding file area tag,
# it could be just one (string) or multiple (tuple) file area tags
_PDS4_PRODUCT_TO_FILE_AREA_TAGS_MAPPING = {
    'Product_Ancillary': ('File_Area_Ancillary',),
    'Product_Browse': ('File_Area_Browse',),
    'Product_Bundle': ('File_Area_Text',),
    'Product_Collection': ('File_Area_Inventory',),
    'Product_External': ('File_Area_External',),
    'Product_File_Repository': ('File_Area_Binary',),
    'Product_File_Text': ('File_Area_Text',),
    'Product_Metadata_Supplemental': ('File_Area_Metadata',),
    'Product_Native': ('File_Area_Native',),
    'Product_Observational': ('File_Area_Observational',
                              'File_Area_Observational_Supplemental'),
    'Product_Proxy_PDS3': ('File_Area_Binary',),
    'Product_SPICE_Kernel': ('File_Area_SPICE_Kernel',),
    'Product_XML_Schema': ('File_Area_XML_Schema',),
}

# The mapping of a table tag to its corresponding record and field tags
# The key is a table tag, and the value is a tuple of the record and field tags
_PDS4_TABLE_TO_RECORD_FIELD_TAGS_MAPPING = {
    'Inventory': ('Record_Delimited', 'Field_Delimited'),
    'Manifest_SIP_Deep_Archive': ('Record_Delimited', 'Field_Delimited'),
    'Table_Binary': ('Record_Binary', 'Field_Binary'),
    'Table_Character': ('Record_Character', 'Field_Character'),
    'Table_Delimited': ('Record_Delimited', 'Field_Delimited'),
    'Table_Delimited_Source_Product_External': ('Record_Delimited', 'Field_Delimited'),
    'Table_Delimited_Source_Product_Internal': ('Record_Delimited', 'Field_Delimited'),
    'Transfer_Manifest': ('Record_Character', 'Field_Character'),
}

# PDS4 label tags under Special_Constants
_PDS4_SPECIAL_CONSTANTS_TAGS = {
    'error_constant',
    'high_instrument_saturation',
    'high_representation_saturation',
    'invalid_constant',
    'low_instrument_saturation',
    'low_representation_saturation',
    'missing_constant',
    'not_applicable_constant',
    'saturated_constant',
    'unknown_constant',
    # 'valid_maximum',
    # 'valid_minimum'
}

# Delimiter used to separate column values in the same row
# It's encoded by 'UTF-8'
PDS4_FIELD_DELIMITER = {
    'Carriage-Return Line-Feed': b'\r\n',
    'Comma': b',',
    'Horizontal Tab': b'\t',
    'Semicolon': b';',
    'Vertical Bar': b'|'
}

# key: PDS4 data type
# value: a tuple of (self._data_type, self._dtype2, self._scalar_func)
PDS4_CHR_DATA_TYPE_MAPPING = {
    'ASCII_Date_DOY': ('time', 'S', tai_from_iso),
    'ASCII_Date_Time_DOY': ('time', 'S', tai_from_iso),
    'ASCII_Date_Time_DOY_UTC': ('time', 'S', tai_from_iso),
    'ASCII_Date_Time_YMD': ('time', 'S', tai_from_iso),
    'ASCII_Date_Time_YMD_UTC': ('time', 'S', tai_from_iso),
    'ASCII_Date_YMD': ('time', 'S', tai_from_iso),
    'ASCII_Time': ('time', 'S', tai_from_iso),
    'ASCII_Integer': ('int', 'int', int),
    'ASCII_NonNegative_Integer': ('int', 'int', int),
    'ASCII_Real': ('float', 'float', float),
    'ASCII_AnyURI': ('string', 'U', None),
    'ASCII_Directory_Path_Name': ('string', 'U', None),
    'ASCII_DOI': ('string', 'U', None),
    'ASCII_File_Name': ('string', 'U', None),
    'ASCII_File_Specification_Name': ('string', 'U', None),
    'ASCII_LID': ('string', 'U', None),
    'ASCII_LIDVID': ('string', 'U', None),
    'ASCII_LIDVID_LID': ('string', 'U', None),
    'ASCII_MD5_Checksum': ('string', 'U', None),
    'ASCII_String': ('string', 'U', None),
    'ASCII_Short_String_Collapsed': ('string', 'U', string_collapsed),
    'ASCII_Short_String_Preserved': ('string_preserved', 'U', None),
    'ASCII_Text_Collapsed': ('string', 'U', string_collapsed),
    'ASCII_Text_Preserved': ('string_preserved', 'U', None),
    'ASCII_VID': ('string', 'U', None),
    'UTF8_String': ('string', 'U', None),
    'UTF8_Short_String_Collapsed': ('string', 'U', string_collapsed),
    'UTF8_Short_String_Preserved': ('string_preserved', 'U', None),
    'ASCII_Boolean': ('boolean', 'bool', None),
    'ASCII_Numeric_Base2': ('int', 'int', int_from_base2),
    'ASCII_Numeric_Base8': ('int', 'int', int_from_base8),
    'ASCII_Numeric_Base16': ('int', 'int', int_from_base16),
}


################################################################################
# Class Pds4TableInfo
################################################################################

class Pds4TableInfo(PdsTableInfo):
    """The Pds4TableInfo class holds the attributes of a PDS4-labeled table."""

    def __init__(self, label_file_path, *, label_contents=None, invalid=None,
                 valid_ranges=None, table_file=None):
        """Load a PDS4 table based on its associated label file.

        Parameters:
            label_file_path (str or Path or FCPath): Path to the PDS4 label file. Even if
                the label contents is provided in `label_list`, `label_file_path` must be
                be specificed because it is needed to locate the table file. If this
                parameter is an FCPath object, it will be downloaded from the remote
                source.
            label_contents (pds4_tools.Label or dict, optional): An option to override
                the parsing of the label. If this is a dict, it is treated as the complete
                structure of the parsed label. If it is a pds4_tools.Label object, it is
                assumed to be the label file that was already parsed. If None, the label
                file specified by `label_file_path` is read and parsed.
            invalid (dict, optional): An optional dictionary keyed by column name. The
                returned value must be a list or set of values that are to be treated as
                invalid, missing, or unknown.
            valid_ranges (dict, optional): An optional dictionary keyed by column name.
                The returned value must be a tuple or list containing the minimum and
                maximum numeric values in that column.
            table_file (str or int, optional): Specify a table file name to be read or an
                integer (1-based) representing the order in which the table appears in the
                label file. A string is treated as a regular expression. If the provided
                table name doesn't exist in the label or the integer is out of the range,
                an error will be raised.
        """

        super().__init__(label_file_path)

        if invalid is None:
            invalid = {}
        if valid_ranges is None:
            valid_ranges = {}

        # Parse the label
        if isinstance(label_contents, Label):
            self._label = label_contents.to_dict()
        elif isinstance(label_contents, dict):
            self._label = label_contents
        elif label_contents:
            raise TypeError('label_contents must be a pds4_tools.Label object, a ' +
                            'dictionary, or None')
        else:
            lbl = Label.from_file(self._label_file_path)
            self._label = lbl.to_dict()

        # Get the file area (table file) info from the label dictionary
        file_areas = []
        for prod_tag, prod_component in self._label.items():
            if prod_tag not in _PDS4_PRODUCT_TO_FILE_AREA_TAGS_MAPPING:
                continue
            file_area_tags = _PDS4_PRODUCT_TO_FILE_AREA_TAGS_MAPPING[prod_tag]
            for current_file_area_tag, current_file_area in prod_component.items():
                if current_file_area_tag not in file_area_tags:
                    continue
                if not isinstance(current_file_area, list):
                    current_file_area = [current_file_area]
                file_areas += current_file_area

        self._table_file_name = None

        if len(file_areas) == 0:
            raise ValueError(f'{label_file_path} does not contain any table file info.')

        try:
            table_name_li = [f['File']['file_name'] for f in file_areas]
        except KeyError:
            raise ValueError('"File/file_name" element not found in PDS4 label for '
                             'one of the File_Area classes')

        if table_file is None:
            if len(file_areas) > 1:
                raise ValueError('The table_file parameter was not specified; it is '
                                 'required because the label contains '
                                 f'{len(table_name_li)} table files: '
                                 f'{", ".join(table_name_li)}')
            else:
                self._table_file_name = table_name_li[0]
                file_area = file_areas[0]
        elif isinstance(table_file, str):
            for idx, name in enumerate(table_name_li):
                if re.match(table_file, name):
                    self._table_file_name = name
                    file_area = file_areas[idx]
                    break
            else:
                raise ValueError(f'The requested table file name "{table_file}" '
                                 'doesn\'t exist. '
                                 f'The label contains {len(table_name_li)} table '
                                 f'files: {", ".join(table_name_li)}')
        elif isinstance(table_file, int):
            if not 1 <= table_file <= len(table_name_li):
                raise ValueError(f'The table_file parameter ({table_file}) is out of the '
                                 f'valid range 1 to {len(table_name_li)}')
            else:
                self._table_file_name = table_name_li[table_file - 1]
                file_area = file_areas[table_file - 1]
        else:
            raise TypeError('table_file must be a string or integer')

        try:
            self._header_bytes = int(file_area['Header']['object_length'])
        except KeyError:
            # Some tables don't have a header
            self._header_bytes = 0

        # Get the table/record/field info by searching the tags in the file area
        table_tag = None
        table_area = None
        record_area = None
        columns = None
        for table_type, (record_tag, field_tag) in \
                _PDS4_TABLE_TO_RECORD_FIELD_TAGS_MAPPING.items():
            # There will only be one table type in the file area so the search order
            # doesn't matter
            if table_type in file_area:
                table_tag = table_type
                table_area = file_area[table_type]
                record_area = table_area.get(record_tag, None)
                columns = record_area.get(field_tag, None)
                break

        if table_area is None:
            raise ValueError(f'No Table type found for "{self._table_file_name}" in '
                             f'"{label_file_path}"; this probably means the File type '
                             'is currently not supported by PdsTable')
        if record_area is None:
            raise ValueError(f'No Record element found for "{self._table_file_name}" in '
                             f'"{label_file_path}"')
        if columns is None:
            raise ValueError(f'No Field element found for "{self._table_file_name}" in '
                             f'"{label_file_path}"')
        if table_tag == 'Table_Binary':
            raise ValueError('Binary table is not supported for '
                             f'"{self._table_file_name}" in "{label_file_path}"')

        try:
            self._rows = int(table_area['records'])
        except (KeyError, ValueError):
            raise ValueError('Missing or bad "records" element for '
                             f'"{self._table_file_name}" in "{label_file_path}"')
        try:
            self._columns = int(record_area['fields'])
        except (KeyError, ValueError):
            raise ValueError('Missing or bad "fields" element for '
                             f'"{self._table_file_name}" in "{label_file_path}"')

        try:
            # for a table with fixed row length
            self._row_bytes = int(record_area['record_length'])
            self._fixed_length_row = True
            self._field_delimiter = None
        except KeyError:
            # for a non-fixed-length table, the row length is not used
            try:
                self._row_bytes = int(record_area['maximum_record_length'])
            except KeyError:
                self._row_bytes = None
            self._fixed_length_row = False
            self._field_delimiter = PDS4_FIELD_DELIMITER[table_area['field_delimiter']]

        # Save the key info about each column in a list and a dictionary
        self._column_info_list = []
        self._column_info_dict = {}

        # Construct the dtype0 dictionary
        self._dtype0 = {'crlf': ('|S2', self._row_bytes-2)}

        default_invalid = set(invalid.get('default', []))

        # Check all the column names, append the suffix _{num} to the duplicated names
        colname = defaultdict(list)
        for idx, col in enumerate(columns):
            name = col['name']
            colname[name].append(idx)

        for name, idx_li in colname.items():
            # append _{num} if there are duplicated names
            if len(idx_li) > 1:
                for num, i in enumerate(idx_li):
                    columns[i]['name'] += f'_{num+1}'

        for col in columns:
            name = col['name']
            field_num = int(col['field_number'])

            pdscol = Pds4ColumnInfo(col, field_num,
                                    invalid=invalid.get(name, default_invalid),
                                    valid_range=valid_ranges.get(name, None))

            self._column_info_list.append(pdscol)
            self._column_info_dict[pdscol.name] = pdscol
            self._dtype0[pdscol.name] = pdscol.dtype0

        table_file_remote_path = (self._label_file_remote_path
                                  .with_name(self._table_file_name))
        self._table_file_path = table_file_remote_path.retrieve()


################################################################################
# class Pds4ColumnInfo
################################################################################

class Pds4ColumnInfo(PdsColumnInfo):
    """The Pds4ColumnInfo class holds the attributes of one column in a PDS4 label."""

    def __init__(self, node_dict, column_no, *, invalid=None, valid_range=None):
        """Constructor for a Pds4ColumnInfo.

        Parameters:
            node_dict (dict): The dictionary associated with the column info obtained
                from a pds4_tools.Label object.
            column_no (int): The index number of this column, starting at zero.
            invalid (set, optional): An optional set of discrete values that are to be
                treated as invalid, missing or unknown.
            valid_range (tuple or list, optional): An optional tuple or list identifying
                the lower and upper limits of the valid range for a numeric column.
        """

        super().__init__()

        if invalid is None:
            invalid = set()

        self._name = node_dict['name']
        self._colno = column_no

        try:
            self._start_byte = int(node_dict['field_location'])
            self._bytes      = int(node_dict['field_length'])
        except KeyError:
            # For a .csv table, each column length is not fixed (and the row is not
            # fixed), so we don't have these values.
            self._start_byte = None
            self._bytes = None

        if self._start_byte is not None and self._bytes is not None:
            self._items = 1
            # Define dtype0 to isolate each column in a record
            self._dtype0 = ('S' + str(self._bytes), self._start_byte - 1)
            self._dtype1 = None  # Multi-item columns are not used in PDS4
        else:
            self._dtype0 = None
            self._dtype1 = None  # Multi-item columns are not used in PDS4

        # Define dtype2 as the intended dtype of the values in the column
        self._data_type = node_dict['data_type']
        # Convert PDS4 data_type
        try:
            (self._data_type,
             self._dtype2,
             self._scalar_func) = PDS4_CHR_DATA_TYPE_MAPPING[self._data_type]
        except KeyError:
            raise ValueError('unsupported data type: ' + self._data_type)

        # Identify validity criteria
        invalid_set = set()
        if valid_range is not None:
            self._valid_range = valid_range
        else:
            valid_max = None
            valid_min = None
            # Search for the 'Special_Constants' tag. If it exists, get the invalid values
            # from tags in PDS4_SPECIAL_CONSTANTS_TAGS and store them in invalid_set.
            if 'Special_Constants' in node_dict:
                special_const_area = node_dict['Special_Constants']
                for invalid_tag in _PDS4_SPECIAL_CONSTANTS_TAGS:
                    invalid_val = special_const_area.get(invalid_tag, None)
                    if invalid_val:
                        if self._scalar_func:
                            try:
                                invalid_val = self._scalar_func(invalid_val)
                            except ValueError:
                                # if the invalid value can't be converted, we will keep
                                # its original value and data type
                                invalid_val = invalid_val

                        invalid_set.add(invalid_val)

                valid_max = special_const_area.get('valid_maximum', None)
                valid_min = special_const_area.get('valid_minimum', None)
                if self._scalar_func:
                    if valid_max is not None:
                        valid_max = self._scalar_func(valid_max)
                    if valid_min is not None:
                        valid_min = self._scalar_func(valid_min)

            if valid_min is not None and valid_max is not None:
                self._valid_range = (valid_min, valid_max)
            else:
                self._valid_range = None

        if isinstance(invalid, (numbers.Real,) + STRING_TYPES):
            invalid_set.add(invalid)
        else:
            invalid_set |= invalid

        self._invalid_values = invalid_set



################################################################################
# UNIT TESTS
################################################################################

from pathlib import Path
import unittest

import numpy as np

from pdstable import PdsTable


class Test_Pds4Table(unittest.TestCase):

    def runTest(self):

        INDEX_PATH = 'test_files/uranus_occultation_ring_fit_rfrench_20201201.xml'
        #######################################################################
        # Test csv table
        #######################################################################
        # CSV_TABLE_FILE_NAME =
        #   'uranus_occultation_ring_fit_rfrench_input_stars_20201201.csv'
        CSV_TABLE_FILE_ORDER = 6
        test_csv_table_basic = PdsTable(label_file=INDEX_PATH,
                                        table_file=CSV_TABLE_FILE_ORDER)

        cwd = Path.cwd()

        self.assertEqual(test_csv_table_basic.label_file_name,
                         INDEX_PATH.split('/')[-1])
        self.assertEqual(test_csv_table_basic.label_file_path, cwd / INDEX_PATH)
        self.assertEqual(test_csv_table_basic.table_file_name,
                         'uranus_occultation_ring_fit_rfrench_input_stars_20201201.csv')
        self.assertEqual(test_csv_table_basic.table_file_path,
                         cwd /
                         ('test_files/uranus_occultation_ring_fit_rfrench_'
                          'input_stars_20201201.csv'))
        # Test strings
        test_star_names = test_csv_table_basic.column_values['Star Name']
        star_name_test_set = np.array(['Bper', 'SSgr', 'U0', 'U0201'])
        self.assertTrue(np.all(star_name_test_set == test_star_names[:4]))

        # Test ints
        test_star_num = test_csv_table_basic.column_values['Star Number']
        star_num_test_set = np.array([3, 8, 12, 16])
        self.assertTrue(np.all(star_num_test_set == test_star_num[:4]))

        # Test floats
        test_ra = test_csv_table_basic.column_values['RA(ICRS)']
        ra_test_set = np.array([47.04220716,
                                283.8163196 ,
                                219.5492129 ,
                                330.1143053])
        self.assertTrue(np.all(ra_test_set == test_ra[:4]))

        # Test dicts_by_row()
        rowdict = test_csv_table_basic.dicts_by_row()
        for i in range(4):
            self.assertEqual(rowdict[i]['RA(ICRS)'], ra_test_set[i])

        #########################################################################
        # Test table_file pointing to a file not a table
        #########################################################################
        error_msg = 'No Table type found for'
        try:
            _ = PdsTable(label_file=INDEX_PATH, table_file=2)
        except ValueError as e:
            self.assertIn(error_msg, str(e),
                          f'"{error_msg}" NOT in error messages: "{str(e)}"')

        try:
            _ = PdsTable(label_file=INDEX_PATH,
                         table_file='uranus_occultation_ring_fit_rfrench_20201201.txt')
        except ValueError as e:
            self.assertIn(error_msg, str(e),
                          f'"{error_msg}" NOT in error messages: "{str(e)}"')

        #########################################################################
        # Test PdsTable instantiation without specifying a valid table name if
        # multiple tables are available
        #########################################################################
        table_files = (
            "uranus_occultation_ring_fit_rfrench_20201201.tab, "
            "uranus_occultation_ring_fit_rfrench_20201201.txt, "
            "uranus_occultation_ring_fit_rfrench_input_data_20201201.tab, "
            "uranus_occultation_ring_fit_rfrench_input_events_20201201.tab, "
            "uranus_occultation_ring_fit_rfrench_input_observatories_20201201.tab, "
            "uranus_occultation_ring_fit_rfrench_input_stars_20201201.csv"
        )
        try:
            test_csv_table_basic = PdsTable(label_file=INDEX_PATH)
        except ValueError as e:
            self.assertIn(table_files, str(e),
                          f'"{table_files}" NOT in error messages: "{str(e)}"')

        try:
            test_csv_table_basic = PdsTable(label_file=INDEX_PATH, table_file='xxx')
        except ValueError as e:
            self.assertIn(table_files, str(e),
                          f'"{table_files}" NOT in error messages: "{str(e)}"')

        ########################################################################
        # Row lookups
        # No File Specification or Bundle Name in .csv table, so return -1
        ########################################################################
        # File Specification
        self.assertEqual(test_csv_table_basic.filespec_column_index(), -1)
        # Bundle Name
        self.assertEqual(test_csv_table_basic.volume_column_index(), -1)

        ########################################################################
        # Row ranges
        # Can't specify row range since rows are not fixed length
        ########################################################################
        error_msg = 'Cannot specify row range for the table without fixed length rows'
        try:
            partial_table = PdsTable(label_file=INDEX_PATH,
                                     row_range=(2, 4),
                                     table_file=CSV_TABLE_FILE_ORDER)
        except ValueError as e:
            self.assertIn(error_msg, str(e),
                          f'"{error_msg}" NOT in error messages: "{str(e)}"')

        # PDS4 TODO: Add tests for invalids & replacements

        #######################################################################
        # Test tab table
        #######################################################################
        TAB_TABLE_FILE_NAME = 'uranus_occultation_ring_fit_rfrench_20201201.tab'
        test_tab_table_basic = PdsTable(label_file=INDEX_PATH,
                                        table_file=TAB_TABLE_FILE_NAME)

        # Test strings
        test_ring_names = test_tab_table_basic.column_values['Ring name']
        ring_name_test_set = np.array(['six', 'five', 'four', 'alpha'])
        self.assertTrue(np.all(ring_name_test_set == test_ring_names[0:4]))

        # Test ints
        test_wavenum = test_tab_table_basic.column_values['Wavenumber']
        wavenum_test_set = np.array([-999, -999, -999, -999])
        self.assertTrue(np.all(wavenum_test_set == test_wavenum[:4]))

        # Test floats
        test_semimajor_axis = test_tab_table_basic.column_values['Semimajor axis']
        semimajor_axis_test_set = np.array([4.1837319048797E+04,
                                            4.2235094301041E+04,
                                            4.2571302273527E+04,
                                            4.4718670266706E+04])

        self.assertTrue(np.all(semimajor_axis_test_set == test_semimajor_axis[:4]))

        # Test dicts_by_row()
        rowdict = test_tab_table_basic.dicts_by_row()
        for i in range(4):
            self.assertEqual(rowdict[i]['Semimajor axis'], semimajor_axis_test_set[i])

        ########################################################################
        # Row lookups, no file spec or bundle name in this table, so return -1
        ########################################################################
        # File Specification
        self.assertEqual(test_tab_table_basic.filespec_column_index(), -1)
        # Bundle Name
        self.assertEqual(test_tab_table_basic.volume_column_index(), -1)

        ####################################
        # Row ranges
        ####################################

        partial_table = PdsTable(label_file=INDEX_PATH,
                                 row_range=(2, 4),
                                 table_file=TAB_TABLE_FILE_NAME)
        self.assertEqual(partial_table.rows, 2)

        self.assertEqual(partial_table.filespec_column_index(), -1)
        self.assertEqual(partial_table.volume_column_index(), -1)

        self.assertEqual(partial_table.find_row_index(**{'Ring name': 'four'}), 0)
        self.assertEqual(partial_table.find_row_index(**{'Ring name': 'alpha'}), 1)

        ####################################
        # PdsLabel input option
        ####################################
        # For PDS4, we store the label dictionary instead of pdsparser.PdsLabel
        # in ._label; therefore we use "==" here instead of "is"
        test = PdsTable(label_file=INDEX_PATH,
                        label_contents=partial_table.pdslabel,
                        table_file=TAB_TABLE_FILE_NAME)
        self.assertTrue(test.pdslabel == partial_table.pdslabel)

        #################################### Invalids
        ####################################
        # A dictionary stores the invalid results of each column at different rows in
        # table: uranus_occultation_ring_fit_rfrench_20201201.tab
        # key is the column name, and value is the number of invalid results in that
        # column.
        cols_with_invalid_results = {
            'Ring name': 0,
            'Semimajor axis': 0,
            'Semimajor axis uncertainty': 0,
            'Eccentricity': 0,
            'Eccentricity uncertainty': 0,
            'Periapse longitude': 0,
            'Periapse uncertainty': 0,
            'Periapse precession rate': 0,
            'Periapse precession rate uncertainty': 1,
            'Periapse precession rate method': 0,
            'Inclination': 0,
            'Inclination uncertainty': 0,
            'Node longitude': 0,
            'Node uncertainty': 0,
            'Nodal regression rate': 0,
            'Nodal regression rate uncertainty': 6,
            'Nodal regression rate method': 0,
            'Wavenumber': 6,
            'Normal mode amplitude': 6,
            'Normal mode amplitude uncertainty': 6,
            'Normal mode phase': 6,
            'Normal mode phase uncertainty': 6,
            'Normal mode pattern speed': 6,
            'Normal mode pattern speed uncertainty': 6,
            'Number of points (Npts)': 0,
            'RMS': 0,
        }

        test_table = PdsTable(label_file=INDEX_PATH, table_file=TAB_TABLE_FILE_NAME)

        rowdict = test_table.dicts_by_row()
        for key in test_table.get_keys():
            if key.endswith('_mask'):
                continue
            rowmasks = test_table.get_column_mask(key)

            if cols_with_invalid_results[key] != 0:
                self.assertTrue(np.any(rowmasks))
            else:
                self.assertFalse(np.any(rowmasks))
            self.assertTrue(isinstance(rowmasks, np.ndarray))

        # rowdict = test_table.dicts_by_row()
        for key in test_table.get_keys():
            if key.endswith('_mask'):
                continue

            rowvals = test_table.get_column(key)
            if np.shape(rowvals[0]) != ():
                continue

            rowmask = test_table.get_column_mask(key)

            if cols_with_invalid_results[key] != 0:
                if rowvals.dtype.kind == 'i':
                    # Wavenumber
                    countv = np.sum(rowvals == -999)
                    self.assertEqual(countv, cols_with_invalid_results[key])
                else:
                    # Rest of columns that have invalid values
                    countv = np.sum(rowvals == -9.99e+99)
                    self.assertEqual(countv, cols_with_invalid_results[key])
            else:
                # columns without invalid values in the table
                countv = 0

            countm = np.sum(rowmask)
            self.assertEqual(countv, countm)

        # 3.0377050292398E-04 appears 4 times in column 'Inclination'
        test_table = PdsTable(label_file=INDEX_PATH,
                              invalid={'default': [3.0377050292398E-04]},
                              table_file=TAB_TABLE_FILE_NAME)
        key = 'Inclination'
        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 4)

        test_table = PdsTable(label_file=INDEX_PATH,
                              invalid={'default': [3.0377050292398E-04],
                                       key: 3.0377050292398E-04},
                              table_file=TAB_TABLE_FILE_NAME)

        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 4)

        ####################################
        # Replacements
        ####################################

        # Replacement as a number
        key = 'Inclination'
        test_table = PdsTable(label_file=INDEX_PATH,
                              invalid={'default': [3.0377050292398E-04, -1.e32]},
                              replacements={key: {3.0377050292398E-04: -1.e32}},
                              table_file=TAB_TABLE_FILE_NAME)

        # replace 3.0377050292398E-04 with -1.e32
        rowvals = test_table.get_column(key)
        self.assertEqual(np.sum(rowvals == 3.0377050292398E-04), 0)
        self.assertEqual(np.sum(rowvals == -1.e32), 4)

        # still 4 invalid values because we put -1.e32 in invalid
        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 4)

        # Replacement as a string
        test_table = PdsTable(label_file=INDEX_PATH,
                              invalid={'default': [3.0377050292398E-04, -1.e32]},
                              replacements={key: {'  3.0377050292398E-04': '  -1.e32'}},
                              table_file=TAB_TABLE_FILE_NAME)

        rowvals = test_table.get_column(key)
        self.assertEqual(np.sum(rowvals == 3.0377050292398E-04), 0)
        self.assertEqual(np.sum(rowvals == -1.e32), 4)

        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 4)

        # Replacement via a callback
        def test_callback_as_str(arg):
            if arg.strip() == '3.0377050292398E-04':
                return '-1e32'
            return arg

        test_table = PdsTable(label_file=INDEX_PATH,
                              invalid={'default': [3.0377050292398E-04, -1.e32]},
                              callbacks={key: test_callback_as_str},
                              table_file=TAB_TABLE_FILE_NAME)

        rowvals = test_table.get_column(key)
        self.assertEqual(np.sum(rowvals == 3.0377050292398E-04), 0)
        self.assertEqual(np.sum(rowvals == -1.e32), 4)

        rowmask = test_table.get_column_mask(key)
        self.assertEqual(np.sum(rowmask), 4)



################################################################################
# pdstable/__init__.py
################################################################################

import os
import warnings

import numpy as np

from .pdsxtable import PdsTableInfo, PdsColumnInfo
from .pds3table import (Pds3TableInfo,
                        PDS3_VOLUME_COLNAMES_lc,
                        PDS3_FILE_SPECIFICATION_COLUMN_NAMES_lc)
from .pds4table import (Pds4TableInfo,
                        PDS4_BUNDLE_COLNAMES_lc,
                        PDS4_FILE_SPECIFICATION_COLUMN_NAMES_lc)
from .utils import is_pds4_label, lowercase_value, tai_from_iso, STRING_TYPES

try:
    from ._version import __version__
except ImportError:  # pragma: no cover
    __version__ = 'Version unspecified'


# This is mainly to make the documentation look good
__all__ = ['PdsTable', 'PdsTableInfo', 'PdsColumnInfo']


class PdsTable:
    """The PdsTable class holds the contents of a PDS-labeled table.

    It is represented by a list of Numpy arrays, one for each column.

    Current limitations for PDS3:
        (1) ASCII tables only, no binary formats.
        (2) Detached PDS labels only.
        (3) Only one data file per label.
        (4) No row or record offsets in the label's pointer to the table file.
        (5) STRUCTURE fields in the label are not supported.
        (6) Columns containing multiple items are not loaded.
        (7) Time fields are represented as character strings unless explicitly
            listed for conversion.
    """
    def __init__(self, label_file, *, label_contents=None, times=None, columns=None,
                       nostrip=None, callbacks=None, ascii=False, replacements=None,
                       invalid=None, valid_ranges=None, table_callback=None,
                       merge_masks=False, filename_keylen=0, row_range=None,
                       table_file=None, label_method='strict'):
        """Constructor for a PdsTable object.

        Parameters:
            label_file (str or Path or FCPath): The path to the PDS label of the table
                file. Must be supplied to get proper relative path resolution.
            label_contents (list or Pds3Label, optional): The contents of the label as a
                list of strings if we shouldn't read it from the file. Alternatively, a
                Pds3Label object to avoid label parsing entirely. Note: this param is for
                PDS3 labels only; it is ignored for PDS4.
            columns (list, optional): An optional list of the names of the columns to
                return. If the list is empty, then every column is returned.
            times (list, optional): An optional list of the names of time columns to be
                stored as floats in units of seconds TAI rather than as strings.
            nostrip (list, optional): An optional list of the names of string columns that
                are not to be stripped of surrounding whitespace.
            callbacks (dict, optional): An optional dictionary that returns a callback
                function given the name of a column. If a callback is provided for any
                column, then the function is called on the string value of that column
                before it is parsed. This can be used to update known syntax errors in a
                particular table.
            ascii (bool, optional): True to interpret the callbacks as translating
                ASCII byte strings; False to interpret them as translating the default str
                type (Unicode).
            replacements (dict, optional): An optional dictionary that returns a
                replacement dictionary given the name of a column. If a replacement
                dictionary is provided for any column, then any value in that column (as a
                string or as its native value) that matches a key in the dictionary is
                replaced by the value resulting from the dictionary lookup.
            invalid (dict, optional): An optional dictionary keyed by column name. The
                returned value must be a list or set of values that are to be treated as
                invalid, missing, or unknown. An optional entry keyed by "default" can be
                a list or set of values that are invalid by default; these are used for
                any column whose name does not appear as a key in the dictionary.
            valid_ranges (dict, optional): An optional dictionary keyed by column name.
                The returned value must be a tuple or list containing the minimum and
                maximum numeric values in that column.
            table_callback (callable, optional): An optional function to be called after
                reading the data table contents before processing them. Note that this
                callback must handle bytestrings.
            merge_masks (bool, optional): True to return a single mask value for each
                column, regardless of how many items might be in that column. False to
                return a separate mask value for each value in a column.
            filename_keylen (int, optional): Number of characters in the filename to use
                as the key of the index if this table is to be indexed by filename. Zero
                to use the entire file basename after stripping off the extension.
            row_range (tuple or list, optional): A tuple or list of integers containing
                the index of the first row to read and the first row to omit. If not
                specified, then all the rows are read.
            table_file (str or int, optional): Specify a table file name to be read or an
                integer (1-based) representing the order in which the table appears in the
                label file. If the provided table name doesn't exist in the label or the
                integer is out of the range, an error will be raised. Only relevant for
                PDS4 labels.
            label_method (str, optional): The method to use to parse the label. Valid
                values are 'strict' (default) or 'fast'. The 'fast' method is faster but
                may not be as accurate. Only relevant for PDS3 labels.

        Notes:
            If both a replacement and a callback are provided for the same column, the
            callback is applied first. The invalid and valid_ranges parameters are applied
            afterward.

            Note that performance will be slightly faster if ascii=True.
        """

        if times is None:
            times = []
        if columns is None:
            columns = []
        if nostrip is None:
            nostrip = []
        if callbacks is None:
            callbacks = {}
        if replacements is None:
            replacements = {}
        if invalid is None:
            invalid = {}
        if valid_ranges is None:
            valid_ranges = {}

        self._is_pds4_lbl = is_pds4_label(label_file)

        # Parse the label
        if self._is_pds4_lbl:
            self._info = Pds4TableInfo(label_file, invalid=invalid,
                                       valid_ranges=valid_ranges,
                                       table_file=table_file)
            self._encoding = {'encoding': 'utf-8'}
        else:
            if table_file is not None:
                raise ValueError('table_file is not supported for PDS3 labels')
            self._info = Pds3TableInfo(label_file, label_contents=label_contents,
                                       invalid=invalid, valid_ranges=valid_ranges,
                                       label_method=label_method)
            self._encoding = {'encoding': 'latin-1'}

        # Select the columns
        if len(columns) == 0:
            self._keys = [info.name for info in self._info.column_info_list]
        else:
            self._keys = columns
        # self._keys is an ordered list containing the name of every column to be
        # returned

        self._keys_lc = [k.lower() for k in self._keys]

        # Load the table data in binary
        if row_range is None:
            self._first = 0
            self._rows = self._info.rows

            with open(self._info.table_file_path, 'rb') as f:
                # Skip over the header
                if self._info.header_bytes != 0:
                    f.seek(self._info.header_bytes)
                lines = f.readlines()

            # Check line count
            if len(lines) != self._info.rows:
                raise ValueError(f'row count mismatch in {label_file}: ' +
                                 f'{len(lines)} rows in file; ' +
                                 f'label says {self._info.rows} rows')

        else:
            if not self._info.fixed_length_row:
                raise ValueError('Cannot specify row range for the table '
                                 'without fixed length rows.')

            self._first = row_range[0]
            self._rows = row_range[1] - row_range[0]
            if self._rows <= 0:
                raise ValueError('row_range must have at least one row')

            header_bytes = self._info.header_bytes
            row_bytes = self._info.row_bytes

            with open(self._info.table_file_path, 'rb') as f:
                f.seek(header_bytes + row_range[0] * row_bytes)
                lines = f.readlines(header_bytes + self.rows * row_bytes - 1)

            if len(lines) > self._rows:
                lines = lines[:self._rows]

            if len(lines) != self._rows:
                raise ValueError(f'row count mismatch: {len(lines)} row(s) read; ' +
                                 f'{self._rows} row(s) requested')

        if table_callback is not None:
            lines = table_callback(lines)

        # For table file with fixed length row:
        # table is now a 1-D array in which the ASCII content of each column
        # can be accessed by name. These are bytes, not strings
        if self._info.fixed_length_row:
            table = np.array(lines)
            try:
                table.dtype = np.dtype(self._info.dtype0)
            except ValueError:
                raise ValueError('Error in row description:\n' +
                                 'old dtype = ' + str(table.dtype) +
                                 ';\nnew dtype = ' + str(np.dtype(self._info.dtype0)))
        # For a table file that doesn't have fixed length row, like a .csv file:
        # table is a 2-D array, each row is an array of the column values for the row.
        else:
            table = np.array([np.array(line.split(self._info.field_delimiter))
                              for line in lines])

        # Extract the substring arrays and save in a dictionary...
        self._column_values = {}
        self._column_masks = {}

        for idx, key in enumerate(self._keys):
            column_info = self._info.column_info_dict[key]
            if self._info.fixed_length_row:
                column = table[key]
            else:
                # Use indexing to access the values of a column for all the rows if the
                # table rows are not fixed length
                column = table[:, idx]

            # column is now a 1-D array containing the ASCII content of this
            # column within each row.

            # For multiple items...
            if self._info.fixed_length_row and column_info.items > 1:

                # Replace the column substring with a list of sub-substrings
                column.dtype = np.dtype(column_info.dtype1)

                items = []
                masks = []
                for i in range(column_info.items):
                    item = column[f'item_{i}']
                    items.append(item)
                    masks.append(False)
                # items is now a list containing one 1-D array for each item in
                # this column.

                self._column_values[key] = items
            else:
                self._column_values[key] = [column]

        # Replace each 1-D array of items from ASCII strings to the proper type
        for key in self._keys:
            column_info  = self._info.column_info_dict[key]
            column_items = self._column_values[key]

            data_type = column_info.data_type
            dtype     = column_info.dtype2
            func      = column_info.scalar_func
            callback  = callbacks.get(key, None)
            repdict   = replacements.get(key, {})
            strip     = (key not in nostrip)

            invalid_values = column_info.invalid_values
            valid_range    = column_info.valid_range

            error_count    = 0
            error_example  = None

            # For each item in the column...
            new_column_items = []
            new_column_masks = []
            for items in column_items:

                invalid_mask = np.zeros(len(items), dtype='bool')

                # Apply the callback if any
                if callback:

                    # Convert string to input format for callback
                    if not ascii:
                        items = items.astype('U')

                    # Apply the callback row by row
                    new_items = []
                    for item in items:
                        new_item = callback(item)
                        new_items.append(new_item)

                    items = np.array(new_items)

                # Apply the replacement dictionary if any pairs are strings
                for (before, after) in repdict.items():
                    if not isinstance(before, STRING_TYPES):
                        continue
                    if not isinstance(after,  STRING_TYPES):
                        continue

                    # The file is read as binary, so the replacements have
                    # to be applied as ASCII byte strings

                    if isinstance(before, (str, np.str_)):
                        before = before.encode(**self._encoding)

                    if isinstance(after, (str, np.str_)):
                        after  = after.encode(**self._encoding)

                    # Replace values (suppressing FutureWarning)
                    items = items.astype('S')
                    with warnings.catch_warnings():
                        warnings.simplefilter('ignore')
                        items[items == before] = after

                # Handle the data type...

                # Handle a string
                if data_type == 'string_preserved':
                    items = items.astype('U')  # No strip

                elif data_type == 'string' or (data_type == 'time' and key not in times):
                    items = items.astype('U')

                    if strip:
                        items = [i.strip() for i in items]
                        items = np.array(items)

                # If this is an int, float or time...

                # Try to convert array dtype
                else:
                    try:
                        items = items.astype(dtype)

                        # Apply the replacements for pairs of this type
                        for (before, after) in repdict.items():
                            with warnings.catch_warnings():
                                warnings.simplefilter('ignore')
                                items[items == before] = after

                        # Convert times if necessary
                        if key in times:
                            items = tai_from_iso(items)

                    # If something went wrong, array processing won't work.
                    # Convert to list and process row by row
                    except Exception:
                        # Process row by row
                        new_items = []
                        for k, item in enumerate(items):
                            try:
                                # Translate the item
                                item = func(item)

                                # Apply a possible replacement
                                item = repdict.get(item, item)

                            # If something went wrong...
                            except Exception:
                                invalid_mask[k] = True

                                error_count += 1
                                if not isinstance(item, str):
                                    item = item.decode(**self._encoding)

                                if strip:
                                    item = item.strip()

                                if error_example is None:
                                    error_example = item

                            # Apply validity criteria to this row
                            invalid_mask[k] |= (item in invalid_values)
                            if valid_range:
                                invalid_mask[k] |= (item < valid_range[0])
                                invalid_mask[k] |= (item > valid_range[1])

                            new_items.append(item)

                        items = new_items

                # Determine validity mask if not already done
                if isinstance(items, np.ndarray):
                    for invalid_value in invalid_values:
                        # Hide FutureWarning for comparisons of different types
                        with warnings.catch_warnings():
                            warnings.simplefilter('ignore')
                            invalid_mask |= (items == invalid_value)

                    if valid_range:
                        # Hide FutureWarning for comparisons of different types
                        with warnings.catch_warnings():
                            warnings.simplefilter('ignore')
                            invalid_mask |= (items < valid_range[0])
                            invalid_mask |= (items > valid_range[1])

                new_column_items.append(items)
                new_column_masks.append(invalid_mask)

            # Swap indices for multiple items
            if len(new_column_items) == 1:
                self._column_values[key] = new_column_items[0]
                self._column_masks[key]  = new_column_masks[0]

            else:
                theyre_all_arrays = np.all([isinstance(c, np.ndarray)
                                            for c in new_column_items])

                if theyre_all_arrays:
                    array = np.stack(new_column_items, axis=1)
                    if array.dtype.kind in ('S', 'U'):
                        array = [tuple(x) for x in array]
                    self._column_values[key] = array
                else:
                    self._column_values[key] = list(zip(*new_column_items))

                if merge_masks:
                    self._column_masks[key] = np.any(np.stack(new_column_masks),
                                                     axis=0)
                else:
                    mask_array = np.stack(new_column_masks)
                    self._column_masks[key] = mask_array.swapaxes(0, 1)

            # Report errors as warnings
            if error_count:
                if error_count == 1:
                    template = (f'Illegally formatted {column_info.data_type} ' +
                                f'value in column {column_info.name}: ' +
                                f'{error_example.strip()}')
                else:
                    template = (f'{str(error_count)} illegally formatted ' +
                                f'{column_info.data_type} values in column ' +
                                f'{column_info.name}; first example is ' +
                                f'"{error_example.strip()}"')

                warnings.warn(template)

        # Cache dicts_by_row and other info when first requested
        self._filename_keylen = filename_keylen
        self._dicts_by_row = {}

        self._volume_colname_index   = None
        self._volume_colname         = None
        self._volume_colname_lc      = None

        self._filespec_colname_index = None
        self._filespec_colname       = None
        self._filespec_colname_lc    = None

        self._rows_by_filename = None
        self._filename_keys    = None

    @property
    def pdslabel(self):
        """The label of the table as a Pds3Label for PDS3 or dict for PDS4."""
        return self._info.label

    @property
    def label_file_name(self):
        """The name of the label file (without the path)."""
        return self._info.label_file_name

    @property
    def label_file_path(self):
        """The local path to the label file."""
        return self._info.label_file_path

    @property
    def table_file_name(self):
        """The name of the table file (without the path)."""
        return self._info.table_file_name

    @property
    def table_file_path(self):
        """The local path to the table file."""
        return self._info.table_file_path

    @property
    def is_pds4(self):
        """True if the read label was a PDS4 label, False otherwise."""
        return self._is_pds4_lbl

    @property
    def rows(self):
        """The number of rows that were read."""
        return self._rows

    @property
    def first(self):
        """The index of the first row that was read (0-based)."""
        return self._first

    @property
    def columns(self):
        """The number of columns in the table (possibly as restricted by the
        columns parameter)."""
        return len(self._column_values)

    @property
    def all_columns(self):
        """The number of columns in the table (possibly as restricted by the
        columns parameter)."""
        return self._info.columns

    @property
    def column_values(self):
        """The values of the columns that were read as a dict indexed by column name."""
        return self._column_values

    @property
    def column_masks(self):
        """The masks of the columns that were read as a dict indexed by column name."""
        return self._column_masks

    @property
    def column_info_list(self):
        """The list of PdsColumnInfo objects for the columns in the table.

        This list includes ALL of the columns, not just the ones restricted by
        the columns parameter.
        """
        return self._info.column_info_list

    @property
    def column_info_dict(self):
        """The dict of PdsColumnInfo objects for the columns in the table, keyed by
        the column name.

        This dict includes ALL of the columns, not just the ones restricted by
        the columns parameter.
        """
        return self._info.column_info_dict

    @property
    def header_bytes(self):
        """The number of bytes in the header of the table."""
        return self._info.header_bytes

    @property
    def encoding(self):
        """The encoding of the table file (e.g., 'utf-8' or 'latin-1')."""
        return self._encoding['encoding']

    @property
    def fixed_length_row(self):
        """True if the table has fixed-length rows."""
        return self._info.fixed_length_row

    @property
    def field_delimiter(self):
        """The field delimiter for the table."""
        return self._info.field_delimiter

    @property
    def row_bytes(self):
        """The number of bytes in a single row of the table."""
        return self._info.row_bytes

    @property
    def dtype0(self):
        """The dtype dictionary for the table, keyed by the column name.

        Each value is a tuple of (dtype_string, start_byte) where dtype_string is the
        string representation of the dtype used to isolate the column (e.g., 'S10' for a
        10-character string) and start_byte is the starting byte position of the column in
        a row.
        """
        return self._info.dtype0

    @property
    def info(self):
        """The Pds3/4TableInfo object that holds the attributes of the table.

        DEPRECATED.
        """
        return self._info

    ############################################################################
    # Support for extracting rows and columns
    ############################################################################

    def dicts_by_row(self, lowercase=(False, False)):
        """Returns a list of dictionaries, one for each row in the table.

        Each dictionary contains all of the column values in that particular row.
        The dictionary keys are the column names; append "_mask" to the key to get
        the mask value, which is True if the column value is invalid; False otherwise.

        Parameters:
            lowercase (tuple or bool): A tuple of two booleans. If the first is
                True, then the dictionary is also keyed by column names converted to
                lower case. If the second is True, then keys with "_lower" appended
                return values converted to lower case. If a single boolean is provided,
                it will be duplicated for both parameters.

        Returns:
            list: A list of dictionaries, one for each row in the table.
        """

        # Duplicate the lowercase value if only one is provided
        if isinstance(lowercase, bool):
            lowercase = (lowercase, lowercase)

        # If we already have the needed list of dictionary, return it
        try:
            return self._dicts_by_row[lowercase]
        except KeyError:
            pass

        # For each row...
        row_dicts = []
        for row in range(self.rows):

            # Create and append the dictionary
            row_dict = {}
            for (column_name, items) in self._column_values.items():

                key_set = set([column_name])
                if not self.is_pds4:
                    key_set = set([column_name, column_name.replace(' ', '_')])

                for key in key_set:
                    value = items[row]
                    mask  = self._column_masks[key][row]

                    # Key and value unchanged
                    row_dict[key] = value
                    row_dict[key + '_mask'] = mask

                    # Key in lower case; value unchanged
                    if lowercase[0]:
                        key_lc = key.lower()
                        row_dict[key_lc] = value
                        row_dict[key_lc + '_mask'] = mask

                    # Value in lower case
                    if lowercase[1]:
                        value_lc = lowercase_value(value)

                        row_dict[key + '_lower'] = value_lc
                        if lowercase[0]:
                            row_dict[key_lc + '_lower'] = value_lc

            row_dicts.append(row_dict)

        # Cache results for later re-use
        self._dicts_by_row[lowercase] = row_dicts

        return row_dicts

    def get_column(self, name):
        """Return the values in the specified column as a list.

        Parameters:
            name (str): The name of the column to retrieve.

        Returns:
            list: The values in the specified column.
        """

        return self._column_values[name]

    def get_column_mask(self, name):
        """Return the masks for the specified column as a list.

        Parameters:
            name (str): The name of the column to retrieve masks for.

        Returns:
            list: The masks for the specified column.
        """

        return self._column_masks[name]

    def get_keys(self):
        """Get the list of column names that were actually loaded.

        Returns:
            list: A list of column names.
        """

        return list(self._keys)

    ############################################################################
    # Support for finding rows by specified column values
    ############################################################################

    def find_row_indices(self, lowercase=(False, False), *,
                               limit=None, substrings=None, **params):
        """Find indices of rows where each named parameter equals the specified value.

        Parameters:
            lowercase (tuple or bool): Whether to enable testing of the column name and
                value converted to lower case. This is a tuple of two booleans. If the
                first is True, then we also allow testing of an entry in `params` with a
                ``_lower`` suffix. If the second boolean is True, then such a column also
                converts the value to match lower case. If a single boolean is provided,
                it will be duplicated for both parameters.
            limit (int, optional): If not zero or None, this is the maximum number of
                matching rows that are returned.
            substrings (list, optional): A list of column names for which a match
                occurs if the given parameter value is embedded within the string; an
                exact match is not required.
            **params: Named parameters where each parameter name corresponds to a column
                name and the value is what to search for in that column.

        Returns:
            list: A list of row indices that match the search criteria.
        """

        if substrings is None:
            substrings = []

        dicts_by_row = self.dicts_by_row(lowercase=lowercase)

        # Make a list (key, value, test_substring, mask_key)
        test_info = []
        for (key, match_value) in params.items():
            if key.endswith('_lower'):
                mask_key = key[:-6] + '_mask'
                match_value = lowercase_value(match_value)
                test_substring = (key in substrings or key[:-6] in substrings)
            else:
                mask_key = key + '_mask'
                test_substring = (key in substrings)

            test_info.append((key, match_value, test_substring, mask_key))

        matches = []

        # For each row in the table...
        for k, row_dict in enumerate(dicts_by_row):

            # Assume it's a match
            match = True

            # Apply each test...
            for (key, match_value, test_substring, mask_key) in test_info:

                # Reject all masked values
                if np.any(row_dict[mask_key]):
                    match = False
                    break

                # Test column value(s)
                column_values = row_dict[key]
                if test_substring:
                    if isinstance(column_values, str):
                        failures = [match_value not in column_values]
                    else:
                        failures = [match_value not in c for c in column_values]
                elif isinstance(column_values, (str, int, float)):
                    failures = [match_value != column_values]
                else:
                    failures = [match_value != c for c in column_values]

                if np.any(failures):
                    match = False
                    break

            # If there were no failures, we have a match
            if match:
                matches.append(k)
                if limit and len(matches) >= limit:
                    return matches

        return matches

    def find_row_index(self, lowercase=(False, False), *, substrings=None, **params):
        """Find the first row where each named parameter equals the specified value.

        Parameters:
            lowercase (tuple or bool): Whether to enable testing of the column name and
                value converted to lower case. This is a tuple of two booleans. If the
                first is True, then we also allow testing of an entry in `params` with a
                ``_lower`` suffix. If the second boolean is True, then such a column also
                converts the value to match lower case. If a single boolean is provided,
                it will be duplicated for both parameters.
            substrings (list, optional): A list of column names for which a match
                occurs if the given parameter value is embedded within the string; an
                exact match is not required.
            **params: Named parameters where each parameter name corresponds to a column
                name and the value is what to search for in that column.

        Returns:
            int: The index of the first matching row.

        Raises:
            ValueError: If no matching row is found.
        """

        if substrings is None:
            substrings = []

        matches = self.find_row_indices(lowercase=lowercase, limit=1,
                                        substrings=substrings, **params)

        if matches:
            return matches[0]

        raise ValueError('row not found: ' + str(params))

    def find_rows(self, lowercase=(False, False), **params):
        """Return a list of dicts representing rows where each named parameter equals
        the specified value.

        Parameters:
            lowercase (tuple or bool): Whether to enable testing of the column name and
                value converted to lower case. This is a tuple of two booleans. If the
                first is True, then we also allow testing of an entry in `params` with a
                ``_lower`` suffix. If the second boolean is True, then such a column also
                converts the value to match lower case. If a single boolean is provided,
                it will be duplicated for both parameters.
            **params: Named parameters where each parameter name corresponds to a column
                name and the value is what to search for in that column.

        Returns:
            list: A list of dictionaries representing the matching rows. Each dictionary
            is keyed by column name.
        """

        indices = self.find_row_indices(lowercase=lowercase, **params)
        dicts_by_row = self.dicts_by_row()
        return [dicts_by_row[k] for k in indices]

    def find_row(self, lowercase=(False, False), **params):
        """Return a dict representing the first row where each named parameter
        equals the specified value.

        Parameters:
            lowercase (tuple or bool): Whether to enable testing of the column name and
                value converted to lower case. This is a tuple of two booleans. If the
                first is True, then we also allow testing of an entry in `params` with a
                ``_lower`` suffix. If the second boolean is True, then such a column also
                converts the value to match lower case. If a single boolean is provided,
                it will be duplicated for both parameters.
            **params: Named parameters where each parameter name corresponds to a column
                name and the value is what to search for in that column.

        Returns:
            dict: A dictionary representing the first matching row. The dictionary is
            keyed by column name.

        Raises:
            ValueError: If no matching row is found.
        """

        k = self.find_row_index(lowercase=lowercase, **params)
        dicts_by_row = self.dicts_by_row()
        return dicts_by_row[k]

    ############################################################################
    # Support for finding rows by filename
    ############################################################################

    def filename_key(self, filename):
        """Convert a filename to a key for indexing the rows.

        The key is the basename with the extension removed.

        Parameters:
            filename (str): The filename to convert to a key.

        Returns:
            str: The filename key for indexing.
        """

        basename = os.path.basename(filename)
        key = os.path.splitext(basename)[0]
        if self._filename_keylen and len(key) > self._filename_keylen:
            key = key[:self._filename_keylen]

        return key

    def bundle_column_index(self):
        """Get the index of the column containing volume IDs or bundle names.

        This is an alias for the volume_column_index() method.

        Returns:
            int: The index of the column containing volume IDs or bundle names,
            or -1 if none.
        """

        return self.volume_column_index()

    def volume_column_index(self):
        """Get the index of the column containing volume IDs or bundle names.

        Returns:
            int: The index of the column containing volume IDs or bundle names,
            or -1 if none.
        """

        if self.is_pds4:
            colnames = PDS4_BUNDLE_COLNAMES_lc
        else:
            colnames = PDS3_VOLUME_COLNAMES_lc

        if self._volume_colname_index is None:
            self._volume_colname_index = -1
            self._volume_colname = ''
            self._volume_colname_lc = ''

            for guess in colnames:
                if guess in self._keys_lc:
                    k = self._keys_lc.index(guess)
                    self._volume_colname_index = k
                    self._volume_colname_lc = guess
                    self._volume_colname = self._keys[k]
                    return k

        return self._volume_colname_index

    def filespec_column_index(self):
        """Get the index of the column containing the file specification name.

        For PDS3 tables, this is a column with a name like "file_specification_name".
        PDS4 tables do not have a standard name, so we look for some possible names.

        Returns:
            int: The index of the column containing the file specification name, or -1 if
            none.
        """

        if self.is_pds4:
            colnames = PDS4_FILE_SPECIFICATION_COLUMN_NAMES_lc
        else:
            colnames = PDS3_FILE_SPECIFICATION_COLUMN_NAMES_lc

        if self._filespec_colname_index is None:
            self._filespec_colname_index = -1
            self._filespec_colname = ''
            self._filespec_colname_lc = ''

            for guess in colnames:
                if guess in self._keys_lc:
                    k = self._keys_lc.index(guess)
                    self._filespec_colname_index = k
                    self._filespec_colname_lc = guess
                    self._filespec_colname = self._keys[k]
                    return k

        return self._filespec_colname_index

    def find_row_indices_by_bundle_filespec(self, bundle_name, filespec=None, *,
                                                  limit=None, substring=False):
        """Find the row indices of the table with the specified bundle_name and
        file_specification_name.

        This is an alias for the find_row_indices_by_volume_filespec() method.

        The search is case-insensitive.

        If the table does not contain the bundle name or if the given value of
        bundle_name is blank or not supplied, the search is performed on the filespec
        alone, ignoring the bundle name. Also, if only one argument is specified,
        it is treated as the filespec.

        The search ignores the extension of filespec so it does not matter
        whether the column contains paths to labels or data files. It also works
        in tables that contain columns of file names without directory paths.

        Parameters:
            bundle_name (str): The bundle name to search for.
            filespec (str, optional): The file specification name to search for.
                If None, bundle_name is treated as the filespec.
            limit (int, optional): Maximum number of matching rows to return.
            substring (bool, optional): If True, a match occurs whenever the given
                filespec appears inside what is tabulated in the file, so a complete
                match is not required.

        Returns:
            list: A list of row indices that match the search criteria.
        """

        return self.find_row_indices_by_volume_filespec(bundle_name, filespec,
                                                        limit=limit,
                                                        substring=substring)

    def find_row_indices_by_volume_filespec(self, volume_id, filespec=None, *,
                                                  limit=None, substring=False):
        """Find the row indices of the table with the specified volume_id and
        file_specification_name.

        The search is case-insensitive.

        If the table does not contain the volume ID or if the given value of
        volume_id is blank or not supplied, the search is performed on the filespec
        alone, ignoring the volume ID. Also, if only one argument is specified,
        it is treated as the filespec.

        The search ignores the extension of filespec so it does not matter
        whether the column contains paths to labels or data files. It also works
        in tables that contain columns of file names without directory paths.

        Parameters:
            volume_id (str): The volume ID to search for.
            filespec (str, optional): The file specification name to search for.
                If None, volume_id is treated as the filespec.
            limit (int, optional): Maximum number of matching rows to return.
            substring (bool, optional): If True, a match occurs whenever the given
                filespec appears inside what is tabulated in the file, so a complete
                match is not required.

        Returns:
            list: A list of row indices that match the search criteria.
        """

        dicts_by_row = self.dicts_by_row(lowercase=(True, True))

        if filespec is None:
            filespec = volume_id
            volume_id = ''

        # Find the name of the columns containing the VOLUME_ID and
        # FILE_SPECIFICATION_NAME
        _ = self.volume_column_index()
        _ = self.filespec_column_index()

        if self._volume_colname is None:
            volume_colname = ''
        else:
            volume_colname = self._volume_colname_lc + '_lower'

        if self._filespec_colname_lc is None:
            raise ValueError('FILE SPECIFICATION NAME column not found')
        else:
            filespec_colname = self._filespec_colname_lc + '_lower'

        example = dicts_by_row[0][self._filespec_colname_lc]

        if not self.is_pds4:
            # Convert to VMS format for really old indices
            if '[' in example:
                parts = filespec.split('/')
                filespec = '[' + '.'.join(parts[:-1]) + ']' + parts[-1]

        # Copy the extension of the example
        filespec = os.path.splitext(filespec)[0]
        if not substring:
            ext = os.path.splitext(example)[1]
            filespec += ext

        # OK now search
        volume_id = volume_id.lower()
        filespec = filespec.lower()
        if substring:
            substrings = [filespec_colname]
        else:
            substrings = []
        if volume_colname and volume_id:
            return self.find_row_indices(lowercase=(True, True),
                                         substrings=substrings, limit=limit,
                                         **{filespec_colname: filespec,
                                            volume_colname: volume_id})
        else:
            return self.find_row_indices(lowercase=(True, True),
                                         substrings=substrings, limit=limit,
                                         **{filespec_colname: filespec})

    def find_row_index_by_bundle_filespec(self, bundle_name, filespec=None, *,
                                                substring=False):
        """Find the first row index with the specified bundle_name and
        file_specification_name.

        This is an alias for the find_row_index_by_volume_filespec() method.

        The search is case-insensitive.

        If the table does not contain the bundle name or if the given value of
        bundle_name is blank, the search is performed on the filespec alone,
        ignoring the bundle name. Also, if only one argument is specified, it is
        treated as the filespec.

        The search ignores the extension of filespec so it does not matter
        whether the column contains paths to labels or data files. It also works
        in tables that contain columns of file names without directory paths.

        Parameters:
            bundle_name (str): The bundle name to search for.
            filespec (str, optional): The file specification name to search for.
                If None, bundle_name is treated as the filespec.
            substring (bool, optional): If True, a match occurs whenever the given
                filespec appears inside what is tabulated in the file, so a
                complete match is not required.

        Returns:
            int: The index of the first matching row.

        Raises:
            ValueError: If no matching row is found.
        """

        return self.find_row_index_by_volume_filespec(bundle_name, filespec,
                                                      substring=substring)

    def find_row_index_by_volume_filespec(self, volume_id, filespec=None,
                                                substring=False):
        """Find the first row index with the specified volume_id and
        file_specification_name.

        The search is case-insensitive.

        If the table does not contain the volume ID or if the given value of
        volume_id is blank, the search is performed on the filespec alone,
        ignoring the volume ID. Also, if only one argument is specified, it is
        treated as the filespec.

        The search ignores the extension of filespec so it does not matter
        whether the column contains paths to labels or data files. It also works
        in tables that contain columns of file names without directory paths.

        Parameters:
            volume_id (str): The volume ID to search for.
            filespec (str, optional): The file specification name to search for.
                If None, volume_id is treated as the filespec.
            substring (bool, optional): If True, a match occurs whenever the given
                filespec appears inside what is tabulated in the file, so a
                complete match is not required.

        Returns:
            int: The index of the first matching row.

        Raises:
            ValueError: If no matching row is found.
        """

        indices = self.find_row_indices_by_volume_filespec(volume_id, filespec,
                                                           limit=1,
                                                           substring=substring)
        if indices:
            return indices[0]

        if volume_id and not filespec:
            raise ValueError(f'row not found: filespec={volume_id}; ')
        elif volume_id:
            raise ValueError(f'row not found: volume_id={volume_id}; filespec={filespec}')
        else:
            raise ValueError(f'row not found: filespec={filespec}')

    def find_rows_by_bundle_filespec(self, bundle_name, filespec=None, *,
                                           limit=None, substring=False):
        """Find the rows of the table with the specified bundle_name and
        file_specification_name.

        This is an alias for the find_rows_by_volume_filespec() method.

        The search is case-insensitive.

        If the table does not contain the bundle name or if the given value of
        bundle_name is blank or not supplied, the search is performed on the filespec
        alone, ignoring the bundle name. Also, if only one argument is specified,
        it is treated as the filespec.

        The search ignores the extension of filespec so it does not matter
        whether the column contains paths to labels or data files. It also works
        in tables that contain columns of file names without directory paths.

        If input parameter substring is True, then a match occurs whenever the
        given filespec appears inside what is tabulated in the file, so a
        complete match is not required.

        Parameters:
            bundle_name (str): The bundle name to search for.
            filespec (str, optional): The file specification name to search for.
                If None, bundle_name is treated as the filespec.
            limit (int, optional): Maximum number of matching rows to return.
            substring (bool, optional): If True, a match occurs whenever the given
                filespec appears inside what is tabulated in the file.

        Returns:
            list: A list of dictionaries representing the matching rows.
        """

        return self.find_rows_by_volume_filespec(bundle_name, filespec,
                                                 limit=limit,
                                                 substring=substring)

    def find_rows_by_volume_filespec(self, volume_id, filespec=None, *,
                                           limit=None, substring=False):
        """Find the rows of the table with the specified volume_id and
        file_specification_name.

        The search is case-insensitive.

        If the table does not contain the volume ID or if the given value of
        volume_id is blank, the search is performed on the filespec alone,
        ignoring the volume ID. Also, if only one argument is specified, it is
        treated as the filespec.

        The search ignores the extension of filespec so it does not matter
        whether the column contains paths to labels or data files. It also works
        in tables that contain columns of file names without directory paths.

        If input parameter substring is True, then a match occurs whenever the
        given filespec appears inside what is tabulated in the file, so a
        complete match is not required.

        Parameters:
            volume_id (str): The volume ID to search for.
            filespec (str, optional): The file specification name to search for.
                If None, volume_id is treated as the filespec.
            limit (int, optional): Maximum number of matching rows to return.
            substring (bool, optional): If True, a match occurs whenever the given
                filespec appears inside what is tabulated in the file.

        Returns:
            list: A list of dictionaries representing the matching rows.
        """

        indices = self.find_row_indices_by_volume_filespec(volume_id, filespec,
                                                           limit=limit,
                                                           substring=substring)
        dicts_by_row = self.dicts_by_row()
        return [dicts_by_row[k] for k in indices]

    def find_row_by_bundle_filespec(self, bundle_name, filespec=None,
                                          substring=False):
        """See find_row_by_volume_filespec."""

        return self.find_row_by_volume_filespec(bundle_name, filespec,
                                                substring=substring)

    def find_row_by_volume_filespec(self, volume_id, filespec=None, *,
                                          substring=False):
        """Find the first row of the table with the specified volume_id and
        file_specification_name.

        The search is case-insensitive.

        If the table does not contain the volume ID or if the given value of
        volume_id is blank, the search is performed on the filespec alone,
        ignoring the volume ID. Also, if only one argument is specified, it is
        treated as the filespec.

        The search ignores the extension of filespec so it does not matter
        whether the column contains paths to labels or data files. It also works
        in tables that contain columns of file names without directory paths.

        Parameters:
            volume_id (str): The volume ID to search for.
            filespec (str, optional): The file specification name to search for.
                If None, volume_id is treated as the filespec.
            substring (bool, optional): If True, a match occurs whenever the given
                filespec appears inside what is tabulated in the file, so a
                complete match is not required.

        Returns:
            dict: A dictionary representing the first matching row.

        Raises:
            ValueError: If no matching row is found.
        """

        k = self.find_row_index_by_volume_filespec(volume_id, filespec=filespec,
                                                   substring=substring)
        dicts_by_row = self.dicts_by_row()
        return dicts_by_row[k]

    def index_rows_by_filename_key(self):
        """Create a dictionary of row indices keyed by the file basename associated
        with the row.

        The key has the file extension stripped away and is converted to lower case.
        The result is available in the filename_keys attribute.
        """

        if self._rows_by_filename is not None:
            return

        _ = self.volume_column_index()
        _ = self.filespec_column_index()

        filespecs = self._column_values[self._filespec_colname]
        masks = self._column_masks[self._filespec_colname]

        rows_by_filename = {}
        filename_keys = []
        for row_num in range(len(filespecs)):
            if masks[row_num]:
                continue

            key = self.filename_key(filespecs[row_num])
            key_lc = key.lower()
            if key_lc not in rows_by_filename:
                rows_by_filename[key_lc] = []
                filename_keys.append(key)

            rows_by_filename[key_lc].append(row_num)

        self._rows_by_filename = rows_by_filename
        self._filename_keys = filename_keys

    @property
    def filename_keys(self):
        """The list of filename keys for the table.

        Returns:
            list: A list of filename keys.
        """

        if self._filename_keys is None:
            self.index_rows_by_filename_key()

        return self._filename_keys

    def row_indices_by_filename_key(self, key):
        """Quick lookup of the row indices associated with a filename key.

        Parameters:
            key (str): The filename key to look up.

        Returns:
            list: A list of row indices associated with the filename key.
        """

        # Create the index if necessary
        self.index_rows_by_filename_key()

        return self._rows_by_filename[key.lower()]

    def rows_by_filename_key(self, key):
        """Quick lookup of the rows associated with a filename key.

        Parameters:
            key (str): The filename key to look up.

        Returns:
            list: A list of dictionaries representing the rows associated with the
            filename key.
        """

        # Create the index if necessary
        self.index_rows_by_filename_key()

        indices = self._rows_by_filename[key.lower()]

        dicts_by_row = self.dicts_by_row()
        rows = [dicts_by_row[k] for k in indices]

        return rows



##########################################################################################
# pdstable/utils.py
# Utility functions for pdstable
##########################################################################################
import julian

import numpy as np


# This is an exhaustive tuple of string-like types
STRING_TYPES = (str, bytes, bytearray, np.str_, np.bytes_)

_PDS4_LBL_EXTENSIONS = {'.xml', '.lblx'}


def is_pds4_label(label_name):
    """Check if the given label is a PDS4 label.

    Parameters:
        label_name (str or Path or FCPath): The name of the label file to check.

    Returns:
        bool: True if the label is a PDS4 label, False otherwise.
    """

    label_name = str(label_name)
    for ext in _PDS4_LBL_EXTENSIONS:
        if label_name.endswith(ext):
            return True
    return False


# Needed because the default value of strip is False
def tai_from_iso(string):
    """Convert ISO time string to TAI seconds.

    Parameters:
        string (str): The ISO time string to convert.

    Returns:
        float: The time in TAI seconds.
    """
    return julian.tai_from_iso(string, strip=True)


def int_from_base2(string):
    """Convert a base-2 string to an integer.

    Parameters:
        string (str): The base-2 string to convert.

    Returns:
        int: The integer value.
    """
    return int(string, 2)


def int_from_base8(string):
    """Convert a base-8 string to an integer.

    Parameters:
        string (str): The base-8 string to convert.

    Returns:
        int: The integer value.
    """
    return int(string, 8)


def int_from_base16(string):
    """Convert a base-16 string to an integer.

    Parameters:
        string (str): The base-16 string to convert.

    Returns:
        int: The integer value.
    """
    return int(string, 16)


def string_collapsed(s):
    """Collapse a string by removing whitespace and converting to lowercase.

    Parameters:
        s (str): The string to collapse.

    Returns:
        str: The collapsed string.
    """
    return s.strip()


def lowercase_value(value):
    """Convert a table value to lower case.

    Handles strings and tuples; leaves ints and floats unchanged.

    Parameters:
        value: The value to convert to lowercase. Can be a string, tuple, numpy array, or
        other type.

    Returns:
        The value converted to lowercase where applicable. Strings are converted to
        lowercase, tuples have their string elements converted to lowercase, numpy arrays
        have their string elements converted to lowercase, and other types are returned
        unchanged.
    """

    if isinstance(value, str):
        value_lc = value.lower()
    elif isinstance(value, tuple):
        value_lc = []
        for item in value:
            if isinstance(item, str):
                value_lc.append(item.lower())
            else:
                value_lc.append(item)
        value_lc = tuple(value_lc)
    elif isinstance(value, np.ndarray):
        value_lc = value.copy()
        for k, val in enumerate(value):
            if isinstance(val, STRING_TYPES):
                value_lc[k] = str(val).lower()
    else:
        value_lc = value

    return value_lc



################################################################################
# pdstable/pdsxtable.py
# PdsTableInfo and PdsColumnInfo
# These are the parent classes for Pds3TableInfo and Pds4TableInfo,
# and Pds3ColumnInfo and Pds4ColumnInfo.
################################################################################

from filecache import FCPath


################################################################################
# Class PdsTableInfo
################################################################################

class PdsTableInfo:
    """Class to hold the attributes of a PDS-labeled table.

    Direct access to this class's attributes by the end user is deprecated and only
    supported for backwards compatibility. Use the properties of PdsTable instead.
    """

    def __init__(self, label_file_path):

        self._label_file_remote_path = FCPath(label_file_path)
        self._label_file_path = self._label_file_remote_path.retrieve()
        self._label_file_name = self._label_file_path.name

        self._label = None
        self._table_file_name = None
        self._table_file_path = None
        self._header_bytes = 0
        self._fixed_length_row = False
        self._field_delimiter = None
        self._rows = 0
        self._columns = 0
        self._row_bytes = 0
        self._column_info_list = []
        self._column_info_dict = {}
        self._dtype0 = {}

    @property
    def label(self):
        """The label of the table as a Pds3Label for PDS3 or dict for PDS4."""
        return self._label

    @property
    def label_file_name(self):
        """The name of the label file (without the path)."""
        return self._label_file_name

    @property
    def label_file_path(self):
        """The local path to the label file."""
        return self._label_file_path

    @property
    def table_file_name(self):
        """The name of the table file (without the path)."""
        return self._table_file_name

    @property
    def table_file_path(self):
        """The local path to the table file."""
        return self._table_file_path

    @property
    def header_bytes(self):
        """The number of bytes in the header of the table."""
        return self._header_bytes

    @property
    def fixed_length_row(self):
        """True if the table has fixed-length rows."""
        return self._fixed_length_row

    @property
    def field_delimiter(self):
        """The field delimiter for the table."""
        return self._field_delimiter

    @property
    def rows(self):
        """The number of rows in the table."""
        return self._rows

    @property
    def columns(self):
        """The number of columns in the table."""
        return self._columns

    @property
    def row_bytes(self):
        """The number of bytes in a single row of the table."""
        return self._row_bytes

    @property
    def column_info_list(self):
        """The list of PdsColumnInfo objects for the columns in the table."""
        return self._column_info_list

    @property
    def column_info_dict(self):
        """The dict of PdsColumnInfo objects for the columns in the table, keyed by
        the column name."""
        return self._column_info_dict

    @property
    def dtype0(self):
        """The dtype dictionary for the table, keyed by the column name.

        Each value is a tuple of (dtype_string, start_byte) where dtype_string is the
        string representation of the dtype (e.g., 'S10' for a 10-character string)
        and start_byte is the starting byte position of the column in a row.
        """
        return self._dtype0


################################################################################
# Class PdsColumnInfo
################################################################################

class PdsColumnInfo:
    """Class to hold the attributes of one column in a PDS-labeled table.

    Direct access of this class's attributes by the end user is not generally necessary,
    but is permitted if you want the inner details of each column.
    """

    def __init__(self):
        self._name = None
        self._colno = None
        self._start_byte = None
        self._bytes = None
        self._items = None
        self._item_bytes = None
        self._item_offset = None
        self._data_type = None
        self._dtype0 = None
        self._dtype1 = None
        self._dtype2 = None
        self._scalar_func = None
        self._valid_range = None
        self._invalid_values = None

    @property
    def name(self):
        """The name of the column."""
        return self._name

    @property
    def colno(self):
        """The index number of the column, starting at zero."""
        return self._colno

    @property
    def start_byte(self):
        """The starting byte of the column in the row (1-based)."""
        return self._start_byte

    @property
    def bytes(self):
        """The number of bytes in the column."""
        return self._bytes

    @property
    def items(self):
        """The number of items in the column (PDS3 only)."""
        return self._items

    @property
    def item_bytes(self):
        """The number of bytes in an item of the column (PDS3 only)."""
        return self._item_bytes

    @property
    def item_offset(self):
        """The incremental offset of each item within the column."""
        return self._item_offset

    @property
    def data_type(self):
        """The data type of the column.

        Possible values are 'int', 'float', 'time', and 'string'.
        """
        return self._data_type

    @property
    def dtype0(self):
        """The dtype of the entire column as a string.

        Each value is a tuple of (dtype_string, start_byte) where dtype_string is the
        string representation of the dtype (e.g., 'S10' for a 10-character string)
        and start_byte is the starting byte position of the column in a row.
        """
        return self._dtype0

    @property
    def dtype1(self):
        """The dtype of a multiple-item column.

        If items == 0, this value is None. Otherwise, it is a dict keyed by 'item_0',
        'item_1', etc. with the value being a tuple of the string representation of
        the dtype used to isolate each item as a string (Snnn) and the starting byte
        relative to the beginning of the column for that item.
        """
        return self._dtype1

    @property
    def dtype2(self):
        """The dtype of the column with the actual data type.

        Possible values are 'int', 'float', 'S', and 'U'.
        """
        return self._dtype2

    @property
    def scalar_func(self):
        """The scalar function used to convert the column's string value to its data
        value."""
        return self._scalar_func

    @property
    def valid_range(self):
        """The valid range of the column as a tuple (lower, upper) or None."""
        return self._valid_range

    @property
    def invalid_values(self):
        """The set of invalid value markers for the column.

        If the column's value equals one of these markers, the column value is considered
        invalid.
        """
        return self._invalid_values



################################################################################
# pdstable/pds3table.py
# Pds3TableInfo and Pds3ColumnInfo
################################################################################

import numbers
import warnings

from pdsparser import Pds3Label

from .pdsxtable import PdsColumnInfo, PdsTableInfo
from .utils import tai_from_iso, STRING_TYPES


# A list of possible table column names that store the volume ID.
PDS3_VOLUME_COLNAMES_lc = (
    'volume_id',
    'volume id',
    'volume_name',
    'volume name'
)

# A list of possible table column names (unique) that store the path of the file spec or
# the file name. This will be used to get the index of the column containing file
# specification path or name. The order of the values matters in this variable; we
# would like to look for the file*specification* first and the file*name, and then the
# rest.
PDS3_FILE_SPECIFICATION_COLUMN_NAMES_lc = (
    'file_specification',
    'file specification',
    'file_specification_name',
    'file specification name',
    'file_name',
    'file name',
    'filename',
    'product_id',
    'product id',
    'stsci_group_id'
)


################################################################################
# Class Pds3TableInfo
################################################################################

class Pds3TableInfo(PdsTableInfo):
    """The Pds3TableInfo class holds the attributes of a PDS3-labeled table."""

    def __init__(self, label_file_path, *, label_contents=None, invalid=None,
                       valid_ranges=None, label_method='strict'):
        """Load a PDS table based on its associated label file.

        Parameters:
            label_file_path (str or Path or FCPath): Path to the PDS3 label file. Even if
                the label contents is provided in `label_list`, `label_file_path` must be
                be specificed because it is needed to locate the table file. If this
                parameter is an FCPath object, it will be downloaded from the remote
                source.
            label_contents (list or dict or Pds3Label, optional): An option to override
                the parsing of the label. If this is a list, it is interpreted as
                containing all the records of the PDS label. If it is a dict, it is
                treated as a parsed label keyed by label item. If it is a Pds3Label
                object, it is assumed to be the label file that was already parsed. If
                None, the label file specified by `label_file_path` is read and parsed.
            invalid (dict, optional): An optional dictionary keyed by column name. The
                returned value must be a list or set of values that are to be treated as
                invalid, missing, or unknown.
            valid_ranges (dict, optional): An optional dictionary keyed by column name.
                The returned value must be a tuple or list containing the minimum and
                maximum numeric values in that column.
            label_method (str, optional): The method to use to parse the label. One of:

                * "strict" performs strict parsing, which requires that the label conform
                  to the full PDS3 standard.
                * "loose" is similar to the above, but tolerates some common syntax
                  errors.
                * "compound" is similar to "loose", but it parses a "compound" label,
                  i.e., one that might contain more than one "END" statement. This option
                  is not supported for attached labels.
                * "fast": uses a different parser, which executes ~ 30x fast than the
                  above and handles all the most common aspects of the PDS3 standard.
                  However, it is not guaranteed to provide an accurate parsing under all
                  circumstances.
        """

        super().__init__(label_file_path)

        if invalid is None:
            invalid = {}
        if valid_ranges is None:
            valid_ranges = {}

        self._header_bytes = 0

        # Parse the label
        if isinstance(label_contents, (Pds3Label, dict)):
            self._label = label_contents
        elif label_contents:
            self._label = Pds3Label(label_contents, method=label_method)
        else:
            self._label = Pds3Label(label_file_path, method=label_method)

        # Get the basic file info...
        if self._label['RECORD_TYPE'] != 'FIXED_LENGTH':
            raise IOError('PDS table does not contain fixed-length records')

        # PDS3 table always has fixed length rows (for now)
        self._fixed_length_row = True

        # Find the pointer to the table file
        # Confirm that the value is a PdsSimplePointer
        self._table_file_name = None
        for key, value in self._label.items():
            if key[0] == '^' and key.endswith('TABLE'):
                self._table_file_name = value
                if key + '_OFFSET' in self._label:
                    msg = ('Table file pointer ' + self._label[key + '_fmt'] +
                           ' is not a Simple Pointer and isn\'t fully ' +
                           'supported')
                    warnings.warn(msg)
                break

        if self._table_file_name is None:
            raise IOError('Pointer to a data file was not found in PDS label')

        # Locate the root of the table object
        table_dict = self._label[key[1:]]

        # Save key info about the table
        interchange_format = (table_dict.get('INTERCHANGE_FORMAT')
                              or table_dict.get('INTERCHANGE_FORMAT_1'))
        if not interchange_format:
            raise ValueError('Table interchange format not specified in label')
        if interchange_format != 'ASCII':
            raise IOError('PDS table is not in ASCII format')

        try:
            self._rows = table_dict['ROWS']
            self._columns = table_dict['COLUMNS']
            self._row_bytes = table_dict['ROW_BYTES']
        except KeyError as e:
            raise ValueError(f'Table definition is missing required field: {e}')

        # Save the key info about each column in a list and a dictionary
        self._column_info_list = []
        self._column_info_dict = {}

        # Construct the dtype0 dictionary
        self._dtype0 = {'crlf': ('|S2', self._row_bytes-2)}

        default_invalid = set(invalid.get('default', []))
        counter = 0
        for key, column_dict in table_dict.items():
            if not isinstance(column_dict, dict):
                continue
            if column_dict['OBJECT'] == 'COLUMN':
                name = column_dict['NAME']
                pdscol = Pds3ColumnInfo(column_dict, counter,
                                        invalid=invalid.get(name, default_invalid),
                                        valid_range=valid_ranges.get(name, None))
                counter += 1

                if name in self._column_info_dict:
                    raise ValueError('duplicated column name: ' + name)

                self._column_info_list.append(pdscol)
                self._column_info_dict[pdscol.name] = pdscol
                self._dtype0[pdscol.name] = pdscol.dtype0

        table_file_remote_path = (self._label_file_remote_path
                                  .with_name(self._table_file_name))
        self._table_file_path = table_file_remote_path.retrieve()


################################################################################
# class Pds3ColumnInfo
################################################################################

class Pds3ColumnInfo(PdsColumnInfo):
    """The Pds3ColumnInfo class holds the attributes of one column in a PDS3 label."""

    def __init__(self, node_dict, column_no, *, invalid=None, valid_range=None):
        """Constructor for a Pds3ColumnInfo.

        Parameters:
            node_dict (dict): The dictionary associated with the pdsparser.PdsNode
                object defining the column.
            column_no (int): The index number of this column, starting at zero.
            invalid (set, optional): An optional set of discrete values that are to be
                treated as invalid, missing, or unknown.
            valid_range (tuple or list, optional): An optional tuple or list identifying
                the lower and upper limits of the valid range for a numeric column.
        """

        super().__init__()

        if invalid is None:  # pragma: no cover
            # This can never be reached because we always just pass an empty set
            invalid = set()

        self._name = node_dict['NAME']
        self._colno = column_no

        try:
            self._start_byte = node_dict['START_BYTE']
            self._bytes      = node_dict['BYTES']
        except KeyError as e:
            raise ValueError(f'Column definition is missing required field: {e}')

        self._items = node_dict.get('ITEMS', 1)
        self._item_bytes = node_dict.get('ITEM_BYTES', self._bytes)
        self._item_offset = node_dict.get('ITEM_OFFSET', self._bytes)

        # Define dtype0 to isolate each column in a record
        self._dtype0 = ('S' + str(self._bytes), self._start_byte - 1)

        # Define dtype1 as a list of dtypes needed to isolate each item
        if self._items == 1:
            self._dtype1 = None
        else:
            self._dtype1 = {}
            byte0 = 0
            for i in range(self._items):
                self._dtype1['item_' + str(i)] = ('S' + str(self._item_bytes), byte0)
                byte0 += self._item_offset

        # Define dtype2 as the intended dtype of the values in the column
        self._data_type = node_dict['DATA_TYPE']
        if 'INTEGER' in self._data_type:
            self._data_type = 'int'
            self._dtype2 = 'int'
            self._scalar_func = int
        elif 'REAL' in self._data_type:
            self._data_type = 'float'
            self._dtype2 = 'float'
            self._scalar_func = float
        elif ('TIME' in self._data_type or 'DATE' in self._data_type or
              self._name.endswith('_TIME') or self._name.endswith('_DATE')):
            self._data_type = 'time'
            self._dtype2 = 'S'
            self._scalar_func = tai_from_iso
        elif 'CHAR' in self._data_type:
            self._data_type = 'string'
            self._dtype2 = 'U'
            self._scalar_func = None
        else:
            raise IOError('unsupported data type: ' + self._data_type)

        # Identify validity criteria
        self._valid_range = valid_range or node_dict.get('VALID_RANGE', None)

        if isinstance(invalid, (numbers.Real,) + STRING_TYPES):
            invalid = set([invalid])

        self._invalid_values = set(invalid)

        self._invalid_values.add(node_dict.get('INVALID_CONSTANT'       , None))
        self._invalid_values.add(node_dict.get('MISSING_CONSTANT'       , None))
        self._invalid_values.add(node_dict.get('UNKNOWN_CONSTANT'       , None))
        self._invalid_values.add(node_dict.get('NOT_APPLICABLE_CONSTANT', None))
        self._invalid_values.add(node_dict.get('NULL_CONSTANT'          , None))
        self._invalid_values.add(node_dict.get('INVALID'                , None))
        self._invalid_values.add(node_dict.get('MISSING'                , None))
        self._invalid_values.add(node_dict.get('UNKNOWN'                , None))
        self._invalid_values.add(node_dict.get('NOT_APPLICABLE'         , None))
        self._invalid_values.add(node_dict.get('NULL'                   , None))
        self._invalid_values -= {None}
